'''
Date: 2024-08-31 20:08:07
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-09-01 22:46:04
Description: 
'''

from typing import Dict, List

from nicegui import ui

from lazydock.pml.shader import Shader, ShaderValues

NULL_CHAIN = "''"
    
    
class ShaderPage:
    def __init__(self, app):
        self._app = app
        self._app.ui_update_func.append(self.ui_update_ui)
        # select receptor
        self.ui_cmap_name = None
        self.ui_shader = None
        self.cmap_name = 'coolwarm'
        self.shaders: Dict[str, Shader] = dict()
        self.now_shader: Shader = None
        # vitulazation control
        self.fig = None
        
    def ui_update_ui(self):
        # update shaders' repr and self.shaders
        for name in list(self.shaders.keys()):
            new_name = repr(self.shaders[name])
            if new_name!= name:
                self.shaders[new_name] = self.shaders.pop(name)
                del self.shaders[name]
        # update ui_shader
        self.ui_shader.set_options(list(self.shaders.keys()))
        
    def build_gui(self):
        with ui.splitter(value = 20).classes('w-full h-full') as splitter:
            with splitter.before:
                # shader
                with ui.card().classes('w-full'):
                    with ui.column().classes('w-full'):
                        ui.label('Shader')
                        with ui.row().classes('w-full'):
                            self.ui_cmap_name = ui.input(label = 'cmap', value = self.cmap_name).bind_value_to(self, 'cmap_name').classes('w-2/3')
                            ui.button('create', on_click=self.create_shader).classes('w-1/3')
                        self.ui_shader = ui.select(list(self.shaders.keys()), label = 'select a shader').bind_value_to(self, 'now_shader').classes('w-full')
                # load shader values
                with ui.card().classes('w-full'):
                    with ui.column().classes('w-full'):
                        ui.label('Shader values').classes('w-full')
            # interaction vitualization
            with splitter.after:
                with ui.row():
                    ui.label('Interaction: ')
                    
    def create_shader(self):
        shader = Shader(self.cmap_name)
        self.shaders[repr(shader)] = shader


class LazyPlot:
    def __init__(self, app, _dev_mode: bool = False):
        self._app = app
        
        self.shader_page = ShaderPage(self._app)
        
    def build_gui(self):
        with ui.tabs().classes('w-full').props('align=left active-bg-color=blue') as tabs:
            self.ui_shader_tab = ui.tab('Pose Shader').props('no-caps')
        with ui.tab_panels(tabs, value=self.ui_shader_tab).classes('w-full'):
            with ui.tab_panel(self.ui_shader_tab):
                self.shader_page.build_gui()
        # return self
        return self
            
        
# dev mode
if __name__ in {"__main__", "__mp_main__"}: 
    from pymol import CmdException, api, cmd   
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/LIGAND.pdb', 'ligand')
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'receptor')
    
    from main import GUILauncher
    GUILauncher()
    


# -*- coding: utf-8 -*-                                                                                     
from random import randint

from pymol import cmd
from pymol.cgo import *


#############################################################################
#                                                                            
# drawBoundingBox.py -- Draws a box surrounding a selection 
#
#                                                                            
# AUTHOR: Jason Vertrees                                                   
# DATE  : 2/20/2009                                                          
# NOTES : See comments below.       
# edited by BHM-Bob: change naming, add quiet and _cmd parameters.                                  
#                                                                            
#############################################################################
def draw_bounding_box(selection="(all)", padding=0.0, linewidth=2.0, r=1.0, g=1.0, b=1.0,
                      quiet=1, _cmd = None):     
    """                                                                  
    DESCRIPTION                                                          
            Given selection, draw the bounding box around it.          

    USAGE:
            drawBoundingBox [selection, [padding, [linewidth, [r, [g, b]]]]]

    PARAMETERS:
            selection,              the selection to enboxen.  :-)
                                    defaults to (all)

            padding,                defaults to 0

            linewidth,              width of box lines
                                    defaults to 2.0

            r,                      red color component, valid range is [0.0, 1.0]
                                    defaults to 1.0                               

            g,                      green color component, valid range is [0.0, 1.0]
                                    defaults to 1.0                                 

            b,                      blue color component, valid range is [0.0, 1.0]
                                    defaults to 1.0                                

    RETURNS
            string, the name of the CGO box

    NOTES
            * This function creates a randomly named CGO box that minimally spans the protein. The
            user can specify the width of the lines, the padding and also the color.                            
    """                                                                                                    
    _cmd = _cmd or cmd
    ([minX, minY, minZ],[maxX, maxY, maxZ]) = _cmd.get_extent(selection)

    if not quiet:
        print("Box dimensions (%.2f, %.2f, %.2f)" % (maxX-minX, maxY-minY, maxZ-minZ))

    minX = minX - float(padding)
    minY = minY - float(padding)
    minZ = minZ - float(padding)
    maxX = maxX + float(padding)
    maxY = maxY + float(padding)
    maxZ = maxZ + float(padding)

    if padding != 0 and not quiet:
        print("Box dimensions + padding (%.2f, %.2f, %.2f)" % (maxX-minX, maxY-minY, maxZ-minZ))

    boundingBox = [
        LINEWIDTH, float(linewidth),

        BEGIN, LINES,
        COLOR, float(r), float(g), float(b),

        VERTEX, minX, minY, minZ,       #1
        VERTEX, minX, minY, maxZ,       #2

        VERTEX, minX, maxY, minZ,       #3
        VERTEX, minX, maxY, maxZ,       #4

        VERTEX, maxX, minY, minZ,       #5
        VERTEX, maxX, minY, maxZ,       #6

        VERTEX, maxX, maxY, minZ,       #7
        VERTEX, maxX, maxY, maxZ,       #8


        VERTEX, minX, minY, minZ,       #1
        VERTEX, maxX, minY, minZ,       #5

        VERTEX, minX, maxY, minZ,       #3
        VERTEX, maxX, maxY, minZ,       #7

        VERTEX, minX, maxY, maxZ,       #4
        VERTEX, maxX, maxY, maxZ,       #8

        VERTEX, minX, minY, maxZ,       #2
        VERTEX, maxX, minY, maxZ,       #6


        VERTEX, minX, minY, minZ,       #1
        VERTEX, minX, maxY, minZ,       #3

        VERTEX, maxX, minY, minZ,       #5
        VERTEX, maxX, maxY, minZ,       #7

        VERTEX, minX, minY, maxZ,       #2
        VERTEX, minX, maxY, maxZ,       #4

        VERTEX, maxX, minY, maxZ,       #6
        VERTEX, maxX, maxY, maxZ,       #8

        END
    ]

    boxName = "box_" + str(randint(0,10000))
    while boxName in _cmd.get_names():
        boxName = "box_" + str(randint(0,10000))

    _cmd.load_cgo(boundingBox,boxName)
    return boxName


'''
Date: 2024-09-01 20:33:00
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-09-04 21:26:46
Description: 
'''
import pickle
import socket
import time
from threading import Lock, Thread
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from mbapy_lite.base import get_fmt_time
from pymol import api as pml_api
from pymol import cmd as pml_cmd


class PymolAPI(object):
    """client module for pymol.api and pymol.cmd"""
    def __init__(self, client: 'VClient', api: str = 'cmd') -> None:
        self._client_ = client
        self._api_ = api
    def __getattribute__(self, name: str) -> Any:
        if name in {'_client_', '_api_'}:
            return super().__getattribute__(name)
        return lambda *args, **kwargs: self._client_.send_action(self._api_, name, *args, **kwargs)
        
        
class VServer:
    """run by pymol"""
    def __init__(self, ip: str = 'localhost', port: int = 8085, verbose: bool = False) -> None:
        self.ip = ip
        self.port = int(port)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1 * 1024 * 1024) # set socket recv buffer size to 1MB
        self.server = Thread(target=self._run_server, daemon=True)
        self.server.start()
        self.lock = Lock()
        self._logs: List[str] = []
        self._copied_log_len = 0
        self._verbose = verbose
        
    def copy_logs(self):
        with self.lock:
            return self._logs.copy()
        
    def copy_new_logs(self):
        with self.lock:
            new_logs = self._logs[self._copied_log_len:].copy()
            self._copied_log_len = len(self._logs)
            return new_logs
        
    def _add_log(self, log: str, verbose: bool = False) -> None:
        with self.lock:
            self._logs.append(f'[{get_fmt_time()}] {log}')
            if self._verbose or verbose:
                print(f'[{get_fmt_time()}] {log}')

    def _recvall(self, sock: socket.socket, count: int):
        buf = b''
        while count:
            newbuf = sock.recv(count)
            if not newbuf: return None
            buf += newbuf
            count -= len(newbuf)
        return buf
    
    def _sendall(self, scok: socket.socket, data: bytes) -> None:
        length = len(data)
        scok.sendall(length.to_bytes(4, 'big'))
        scok.sendall(data)
        
    def _run_server(self) -> None:
        """main loop run in a thread"""
        self.socket.bind((self.ip, self.port))
        self.socket.listen(1)
        while True:
            conn, addr = self.socket.accept()
            self._add_log(f'Connected by {addr}', verbose=True)
            while True:
                # get and un-pickle data from client
                recv_data = conn.recv(4)
                if not recv_data:
                    continue
                data_length = int.from_bytes(recv_data, 'big')
                recv_data = self._recvall(conn, data_length)
                if not recv_data:
                    self._add_log(f'VServer: Failed to receive data from client for length {data_length}')
                    continue
                if recv_data == b'quit':
                    self._add_log('VServer: Client quit')
                    break
                api, fn, args, kwargs = pickle.loads(recv_data)
                # execute action
                if api not in {'cmd', 'api'}:
                    self._add_log(f'VServer: Error in executing {api}.{fn}: api not found')
                try:
                    if api == 'cmd':
                        ret = getattr(pml_cmd, fn)(*args, **kwargs)
                    elif api == 'api':
                        ret = getattr(pml_api, fn)(*args, **kwargs)
                    self._sendall(conn, pickle.dumps(ret))
                    self._add_log(f'VServer: {api}.{fn} executed successfully')
                except Exception as e:
                    self._add_log(f'VServer: Error in executing {api}.{fn}: {e}')
                # sleep to avoid busy loop
                time.sleep(0.05)
            # close connection
            conn.close()
            self._add_log(f'Connect closed by {addr}', verbose=True)
        
        
class VClient(VServer):
    """run by user in anthor python script"""
    def __init__(self, ip: str = 'localhost', port: int = 8085) -> None:
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 1 * 1024 * 1024) # set socket send buffer size to 1MB
        self.socket.connect((ip, port))
        
    def _run_server(self) -> None:
        raise NotImplementedError('VClient cannot run as a server now')
    
    def send_action(self, api: str, fn: str, *args, **kwargs) -> Any:
        data = pickle.dumps((api, fn, args, kwargs))
        self._sendall(self.socket, data)
        try:
            ret_len = int.from_bytes(self.socket.recv(4), 'big')
            ret = self._recvall(self.socket, ret_len)
            return pickle.loads(ret)
        except Exception as e:
            print(f'Error in receiving data: {e}')
            return None
        
    def sen_quit(self) -> None:
        data = b'quit'
        self._sendall(self.socket, data)
        self.socket.close()
        
        
__all__ = [
    'PymolAPI',
    'VServer',
    'VClient',
]
        

def _test_server():
    pml_cmd.reinitialize()
    server = VServer()
    while True:
        time.sleep(0.1)
        print(server.copy_new_logs())
        

def _test_client():
    client = VClient()
    vcmd = PymolAPI(client, 'cmd')
    print(vcmd.load('data_tmp/pdb/LIGAND.pdb', 'ligand'))
    

if __name__ == '__main__':
    from multiprocessing import Process
    p1 = Process(target=_test_server)
    p2 = Process(target=_test_client)
    p1.start()
    time.sleep(3)
    p2.start()
    p1.join()
    p2.join()
    



from pathlib import Path
from typing import Dict, List, Union

import pandas as pd
from mbapy_lite.base import put_log
from mbapy_lite.file import opts_file
from mbapy_lite.web import Browser, random_sleep
from pymol import cmd

if __name__ == '__main__':
    from lazydock.pml.thirdparty.draw_bounding_box import draw_bounding_box
    from lazydock.utils import uuid4
else:
    from ..pml.thirdparty.draw_bounding_box import draw_bounding_box
    from ..utils import uuid4


def get_pocket_box_from_ProteinPlus(receptor_path: str, ligand_path: str,
                                    browser: Browser = None):
    """
    download pocket result zip file from ProteinPlus.
    
    Parameters
        - receptor_path: str, path to the receptor pdb file.
        - ligand_path: str, path to the ligand sdf file.
        - browser: lazydock.web.Browser, browser object to use. If None, a new browser will be created.
        
    Returns
        None, the pocket result zip file will be downloaded to the download_path specified in the browser.
    """
    receptor_path = str(Path(receptor_path).resolve())
    ligand_path = str(Path(ligand_path).resolve())
    if not browser:
        download_path = str(Path(receptor_path).resolve().parent)
        b = Browser(download_path=str(Path(receptor_path).resolve().parent))
        put_log(f'using download path: {download_path}')
    else:
        b = browser
    b.get('https://proteins.plus/')
    btn = b.find_elements('//*[@id="pdb_file_pathvar"]')[0]
    btn.send_keys(receptor_path)
    btn = b.find_elements('//*[@id="pdb_file_userligand"]')[0]
    btn.send_keys(ligand_path)
    b.click(element='//*[@id="new_pdb_file"]/input[6]')
    # now in the result page
    b.click(element='//*[@id="headingTwo"]/h4/a/span')
    b.click(element='/html/body/div[4]/div[3]/div/div[1]/div/div/div[2]/div[2]/div/a')
    b.click(element='/html/body/div[4]/div[3]/div/div[3]/div[3]/form/div[8]/div/input')
    while not b.wait_element(element='/html/body/div[4]/div[3]/div/div[3]/div[3]/div[3]/div[1]/form/input[1]'):
        random_sleep(3)
    b.click(element='/html/body/div[4]/div[3]/div/div[3]/div[3]/div[3]/div[1]/form/input[1]')
    # distroy browser
    if not browser:
        b.browser.quit()
    return None

def parse_pocket_box_from_ProteinPlus(result_path: str, k: Union[int, List[int]] = None,
                                      reinitialize: bool = False, draw_box: bool = True,
                                      _cmd = None, method: str = 'extend'):
    """
    parse zip result file from ProteinPlus and return box information.
    
    Parameters
        - result_path: str, path to the zip file downloaded from ProteinPlus.
        - k: int or list of int, index of the pocket to be parsed. If None, all pockets will be parsed.
        - reinitialize: bool, whether to reinitialize pymol or not.
        - draw_box: bool, whether to draw box or not.
        - _cmd: pymol.cmd, pymol command object, can be pml.server.PymolAPI
        - method: str, method to calculate box center and size. 'extend' to use pymol's get_extent function, 'mean' to calculate mean coordinates of all atoms in the pocket.
        
    Returns
        box_df: pandas.DataFrame, dataframe of pocket information. Only return in the method of 'mean'.
            - 'resn_resi_index': str, residue name, residue index and atom index.
            -  'X', 'Y', 'Z': float, coordinates of the atom.
        box_info: dict, box information.
            - 'box_center': list of float, center of the box.
            - 'box_size': list of float, size of the box.
            - 'box_vertices': list of list of float, vertices of the box.
        
    """
    _cmd = _cmd or cmd
    if reinitialize:
        _cmd.reinitialize()
    # load result file
    files = opts_file(result_path, way='zip')
    pdbs = list(filter(lambda x: '_P_' in x and x.endswith('.pdb'), files))
    # check parameter k
    if k is None:
        k = [0]
    elif isinstance(k, int):
        k = [k]
    elif not isinstance(k, list):
        raise TypeError(f'k must be int or list, got {type(k)}.')
    if isinstance(k, list) and any((not any(f'_P_{i}_' in p for p in pdbs)) for i in k):
        raise ValueError(f'k contains invalid index: {k}.')
    # load pocket into pymol
    pocket_name = f'pocket_{uuid4()}'
    for idx, i in enumerate(k):
        pdb_i = list(filter(lambda x: f'_P_{i}_' in x, pdbs))[0]
        _cmd.read_pdbstr(files[pdb_i], f'{pocket_name}_{i}')
        if idx == 0:
            _cmd.select(pocket_name, f'{pocket_name}_{i}')
        else:
            _cmd.select(pocket_name, f'{pocket_name} or {pocket_name}_{i}')
    # draw box
    if draw_box:
        draw_bounding_box(pocket_name, quiet=1, _cmd = _cmd)
    # calcu box
    if method == 'extend':
        ([minX, minY, minZ], [maxX, maxY, maxZ]) = _cmd.get_extent(pocket_name)
        _cmd.delete(pocket_name)
        return {
            'box_center': [(minX + maxX) / 2, (minY + maxY) / 2, (minZ + maxZ) / 2],
            'box_size': [maxX - minX, maxY - minY, maxZ - minZ],
            'box_vertices': [[minX, minY, minZ], [maxX, maxY, maxZ]],
        }
    elif method == 'mean':
        df = pd.DataFrame(columns=['resn_resi_index', 'X', 'Y', 'Z'])
        df.set_index(['resn_resi_index'])
        cmd.iterate_state(1, pocket_name, 'df.loc[f"{resn}_{resi}_{index}", ("X", "Y", "Z")] = [x, y, z]',
                          space=locals())
        _cmd.delete(pocket_name)
        return df, {
            'box_center': df[['X', 'Y', 'Z']].mean().tolist(),
            'box_size': (df[['X', 'Y', 'Z']].max() - df[['X', 'Y', 'Z']].min()).tolist(),
            'box_vertices': [df[['X', 'Y', 'Z']].min().tolist(), df[['X', 'Y', 'Z']].max().tolist()],
        }
    
    
__all__ = [
    'get_pocket_box_from_ProteinPlus',
    'parse_pocket_box_from_ProteinPlus',
]

if __name__ == '__main__':
    # from mbapy_lite.base import Configs
    # b = Browser(options=['--no-sandbox', f"--user-agent={Configs.web.chrome_driver_path}"],
    #             download_path=os.path.abspath('data_tmp/pdb'))
    # get_pocket_box_from_ProteinPlus('data_tmp/pdb/RECEPTOR.pdb', 'data_tmp/pdb/LIGAND.sdf', browser=b)
    df, box = parse_pocket_box_from_ProteinPlus('data_tmp/pdb/POCKET.zip', [1], True, method='mean')
    print(parse_pocket_box_from_ProteinPlus('data_tmp/pdb/POCKET.zip', [1, 2, 3], True))
    print(parse_pocket_box_from_ProteinPlus('data_tmp/pdb/POCKET.zip', [1, 2, 3, 4, 5], True))


# edited from https://github.com/luost26/diffab/blob/main/diffab/tools/relax/pyrosetta_relaxer.py and Kong X, Jia Y, Huang W, et al. Full-Atom Peptide Design with Geometric Latent Diffusion[C],2024.

import pyrosetta
from pyrosetta.rosetta import protocols
from pyrosetta.rosetta.core.pack.task import TaskFactory, operation
from pyrosetta.rosetta.core.scoring import ScoreType
from pyrosetta.rosetta.core.select import residue_selector as selections
from pyrosetta.rosetta.core.select.movemap import (MoveMapFactory,
                                                   move_map_action)
from pyrosetta.rosetta.protocols.relax import FastRelax

init_options = ' '.join([
    '-mute', 'all',
    '-use_input_sc',
    '-ignore_unrecognized_res',
    '-ignore_zero_occupancy', 'false',
    '-load_PDB_components', 'false',
    '-relax:default_repeats', '2',
    '-no_fconfig',
    '-use_terminal_residues', 'true',
    '-in:file:silent_struct_type', 'binary',
])


class RelaxPDBChain:
    def __init__(self, max_iter: int = 1000) :
        self.scorefxn = pyrosetta.create_score_function('ref2015')
        self.fast_relax = FastRelax()
        self.fast_relax.set_scorefxn(self.scorefxn)
        self.fast_relax.max_iter(max_iter)
        
    def __call__(self, pdb_path: str, chain: str):
        original_pose = pyrosetta.pose_from_pdb(pdb_path)
        tf = TaskFactory()
        tf.push_back(operation.InitializeFromCommandline())
        tf.push_back(operation.RestrictToRepacking()) # Only allow residues to repack No design at any position.
        gen_selector = selections.ChainSelector(chain)
        nbr_selector = selections.NeighborhoodResidueSelector()
        nbr_selector.set_focus_selector(gen_selector)
        nbr_selector.set_include_focus_in_subset(True)
        subset_selector = nbr_selector
        prevent_repacking_rlt = operation.PreventRepackingRLT()
        prevent_subset_repacking = operation.OperateOnResidueSubset(prevent_repacking_rlt, subset_selector, flip_subset =True,)
        tf.push_back(prevent_subset_repacking)
        fr = self.fast_relax
        pose = original_pose.clone()
        mmf = MoveMapFactory()
        mmf.add_bb_action(move_map_action.mm_enable, gen_selector)
        mmf.add_chi_action(move_map_action.mm_enable, subset_selector)
        mm = mmf.create_movemap_from_pose(pose)
        fr.set_movemap(mm)
        fr.set_task_factory(tf)
        fr.apply(pose)
        return pose


'''
Date: 2024-08-31 21:40:56
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-09-03 16:52:25
Description: 
'''
from dataclasses import dataclass
from threading import Lock
from typing import Dict, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import Colormap, TwoSlopeNorm
from mbapy_lite.plot import rgb2hex
from pymol import cmd

from lazydock.utils import uuid4


@dataclass
class ShaderAtom:
    model: str
    chain: str
    resi: int
    elem: str
    index: int
    c_value: float = None
    alpha: float = 1.0


@dataclass
class ShaderRes:
    model: str
    chain: str
    resi: int
    atoms: List[ShaderAtom] = None
    c_value: float = None
    alpha: float = 1.0

    def get_c_value(self):
        """
        Returns:
            - c_value, if it is not None.
            - sum of c_values of atoms, if atoms is not None and c_value is None.
        Notes: 
            - If atoms is updated after c_value calculation, c_value will not be updated.
        """
        if self.atoms is not None and self.c_value is None:
            self.c_value = sum(map(lambda x: x.c_value, filter(lambda x: x.c_value, self.atoms)))
        return self.c_value


class ShaderValues:
    def __init__(self, chains: Dict[str, List[ShaderRes]] = None) -> None:
        self.chains: Dict[str, List[ShaderRes]] = chains or {}

    def get_c_value(self, chain: str, resi: int, atom_index: int = None):
        if chain not in self.chains:
            raise ValueError(f'Chain {chain} not found in shader values.')
        res = [res for res in self.chains[chain] if res.resi == resi]
        if not res:
            raise ValueError(f'Residue {resi} not found in chain {chain}.')
        if atom_index is not None:
            atom = [atom for atom in res[0].atoms if atom.index == atom_index]
            if not atom:
                raise ValueError(f'Atom {atom_index} not found in residue {resi}.')
            return atom[0].c_value
        return res[0].get_c_value()
    
    def get_all_c_values(self, level: str = 'res'):
        if level not in {'res', 'atom'}:
            raise ValueError(f'Level must be "res" or "atom", got {level}.')
        if level == 'res':
            return [(res.model, res.chain, res.resi, res.alpha, res.get_c_value()) for chain in self.chains for res in self.chains[chain]]
        else:
            return [(atom.model, atom.chain, atom.resi, atom.index, res.alpha, atom.c_value) for chain in self.chains for res in self.chains[chain] for atom in res.atoms]
          
    def from_interaction_df(self, df: Union[str, pd.DataFrame], model: str, sum_axis: int = 0):
        """
        load values from interaction_df calculated by lazudock.interaction_utils.calcu_receptor_poses_interaction
        
        Parameters:
            df (Union[str, pd.DataFrame]): interaction_df or path to interaction_df.
            model (str): model name.
            sum_axis (int): 0 or 1, pass to df.sum, 0 means sum each rows and get a colum, 1 means sum each colums and get a row.
            
        Returns:
            ShaderValues: self.
        """
        if isinstance(df, str):
            df = pd.read_excel(df, index_col = 0)
        df = df.sum(axis=sum_axis)[1:] # first row is index and colum names
        for res, v in zip(df.index, df.values):
            chain, resi, _ = res.split(':')
            self.chains.setdefault(chain, []).append(ShaderRes(model, chain, int(resi), c_value=v))
        return self


class Shader:
    global_locker = Lock()
    global_name2col = {}
    def __init__(self, cmap: Union[str, Colormap] = 'coolwarm',
                 norm: TwoSlopeNorm = None,
                 col_name_prefix: str = 'COL_') -> None:
        self.cmap = plt.get_cmap(cmap) if isinstance(cmap, str) else cmap
        self.norm = norm or TwoSlopeNorm(vmin=-1, vcenter=0, vmax=1)
        self.COL_NAME_PREFIX = col_name_prefix
        
    def get_col_from_c_value(self, c_value: float):
        """return rgba and color name with prefix for given c_value."""
        rgba = self.cmap(self.norm(c_value))
        name = rgb2hex(*[int(rgba[i]*255) for i in range(3)])
        return rgba, f'{self.COL_NAME_PREFIX}{name[1:]}'
                
    def _get_rgba_col_name(self, c_value: float, _col_name: str = None, _cmd = None):
        """get rgba and color name with prefix for given c_value, 
        store name in self.global_name2col if not exist."""
        _cmd = _cmd or cmd
        rgba, col_name = self.get_col_from_c_value(c_value)
        col_name = _col_name or col_name
        if col_name not in self.global_name2col:
            _cmd.set_color(col_name, list(rgba[:-1]))
            with self.global_locker: # NOTE: only lock-unlock when adding new color to avoid ReLock
                self.global_name2col[col_name] = rgba
        return rgba, col_name
        
    def create_colors_in_pml(self, values: ShaderValues, level: str = 'res',
                             names: List[str] = None, _cmd = None):
        """
        set color and it's name in pymol for each c_value in values.
        
        Parameters:
            values (ShaderValues): values to create colors for.
            level (str): 'atom' or'res', level to create colors for.
            names (List[str]): list of color names to use. If None, will generate names from c_values.
            cmd: pymol cmd module, can be pymol.cmd or lazudock.pml.server.PymolAPI
        Notes:
            If name is given, the color name prefix will not be added.
        """
        c_values = np.unique([v[-1] for v in values.get_all_c_values(level)])
        if isinstance(names, list) and len(names) != len(c_values):
            raise ValueError(f'If names is given, names must be equal to length of c_values, got {len(names)} and {len(c_values)}.')
        for i, c in enumerate(c_values):
            name = names[i] if names is not None else None
            if name not in self.global_name2col:
                self._get_rgba_col_name(c, name, _cmd)
                
    def auto_scale_norm(self, c_values):
        self.norm.autoscale(list(map(lambda x: x[-1], c_values)))
        
                
    def apply_shader_values(self, values: ShaderValues, level: str = 'res',
                            auto_detect_vlim: bool = True, alpha_mode: str = None,
                            _cmd = None):
        """
        apply shader values to pymol.
        
        Parameters:
            values (ShaderValues): values to apply.
            level (str): 'atom' or'res', level to apply values for.
            auto_detect_vlim (bool): if True, will set vmin and vmax for colormap based on min and max c_values.
            alpha_mode (str): pymol transparency mode, such as `cartoon_transparency`.
            _cmd: pymol cmd module, can be pymol.cmd or lazudock.pml.server.PymolAPI
        """
        _cmd = _cmd or cmd
        if level not in {'res', 'atom'}:
            raise ValueError(f'Level must be "res" or "atom", got {level}.')
        c_values = values.get_all_c_values(level)
        if auto_detect_vlim:
            self.auto_scale_norm(c_values)
        # loop through residues or atoms and apply color
        for c_value in c_values:
            if level =='res':
                model, chain, resi, alpha, c = c_value
                _, col_name = self._get_rgba_col_name(c, _cmd=_cmd)
                sele_exp = f'model {model} and (chain {chain} and resi {resi})'
                _cmd.color(col_name, sele_exp)
                if alpha_mode is not None:
                    _cmd.set(alpha_mode, alpha, sele_exp)
            else: # level == 'atom' and res.atoms is not None
                model, chain, resi, atom_index, alpha, c = c_value
                _, col_name = self._get_rgba_col_name(c, _cmd=_cmd)
                sele_exp = f'(model {model} and chain {chain}) and (resi {resi} and index {atom_index})'
                _cmd.color(col_name, sele_exp)
                if alpha_mode is not None:
                    _cmd.set(alpha_mode, alpha, sele_exp)
                    
    def apply_shader_values_to_selection(self, selection, c_value: float = None,
                                         alpha: float = 1.0, alpha_mode: str = None,
                                         _cmd = None):
        _cmd = _cmd or cmd
        if c_value is not None:
            _, col_name = self._get_rgba_col_name(c_value, _cmd=_cmd) # NOTE: do not use alpha from cmap
            _cmd.color(col_name, selection)
        if alpha_mode is not None:
            _cmd.set(alpha_mode, alpha, selection)
                    
    def apply_shader_values_to_sele(self, select_expression: str, c_value: float = None,
                                    alpha: float = 1.0, alpha_mode: str = None,
                                    _cmd = None):
        _cmd = _cmd or cmd
        selection = uuid4()
        _cmd.select(selection, select_expression)
        self.apply_shader_values_to_selection(selection, c_value, alpha, alpha_mode, _cmd)
        _cmd.delete(selection)
                    
    def __repr__(self):
        return f'{self.cmap.name}({self.norm.vmin:.2f},{self.norm.vcenter:.2f},{self.norm.vmax:.2f}){self.COL_NAME_PREFIX}'
    
    
__all__ = [
    'ShaderAtom',
    'ShaderRes',
    'ShaderValues',
    'Shader',
]
    
    
if __name__ == '__main__':
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'receptor')
    values = ShaderValues().from_interaction_df('data_tmp/interaction_df.xlsx', 'receptor')
    shader = Shader()
    shader.create_colors_in_pml(values)
    shader.apply_shader_values(values, alpha_mode='cartoon_transparency')
    print(shader)
    cmd.clip('near', -10)
    cmd.clip('slab', 100)
    cmd.zoom('model receptor')
    cmd.png('data_tmp/shader.png', 1200, 1200, 300)


from typing import List, Union

import pyrosetta
from pyrosetta.rosetta.protocols.analysis import InterfaceAnalyzerMover


def calcu_interface_energy(pdb_path: str, receptor_chains: Union[str, List[str]],
                           ligand_chains: Union[str, List[str]]) -> float:
    """
    use pyrosetta to calculate the interface energy between receptor and ligand
    """
    pose = pyrosetta.pose_from_pdb(pdb_path)
    interface = ''.join(ligand_chains) + '_' + ''.join(receptor_chains)
    mover = InterfaceAnalyzerMover(interface)
    mover.set_pack_separated(True)
    mover.apply(pose)
    return pose.scores['dG_separated']


'''
Date: 2024-09-15 22:05:00
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-10-17 16:38:16
Description: 
'''
import re
import time
from pathlib import Path
from typing import Dict, List, Union

import requests
from lxml import etree
from mbapy_lite.base import put_err
from mbapy_lite.file import opts_file
from mbapy_lite.web import BROWSER_HEAD, Browser, random_sleep


def get_score_from_proq(pdb_path: str, **kwargs) -> Dict[str, float]:
    """return LGscore and MaxSub in dict"""
    session = requests.Session()
    session.headers.update({'User-Agent': BROWSER_HEAD})
    response = session.post('https://proq.bioinfo.se/cgi-bin/ProQ/ProQ.cgi',
                            files = {'pdbfile': opts_file(pdb_path, 'rb')})
    xml = etree.HTML(response.content.decode())
    LGscore = float(xml.xpath('/html/body/b[1]/text()')[0])
    MaxSub = float(xml.xpath('/html/body/b[2]/text()')[0])
    return {'LGscore': LGscore, 'MaxSub': MaxSub}


def get_score_from_VoroMQA(pdb_path: str, browser: Browser = None, timeout: int = 1200, **kwargs) -> Dict[str, float]:
    """return Score in dict"""
    pdb_path = str(Path(pdb_path).resolve())
    b = browser or Browser()
    b.get('https://bioinformatics.lt/wtsam/voromqa/submit')
    btn = b.find_elements('//input[@type="file"]')[0]
    btn.send_keys(pdb_path)
    b.click(element='//button[text()="Submit"]')
    xpath_result_line = '//div[text()="Results:"]/..//div[2]'
    if not b.wait_element(xpath_result_line, timeout=timeout):
        return put_err('Timeout', {})
    score_text = b.find_elements(xpath_result_line)[0].text
    score = float(re.findall(r'Score: ([-\d.]+)', score_text)[0])
    return {'Score': score}


def get_score_from_ProSA(pdb_path: str, browser: Browser = None, timeout: int = 1200, **kwargs) -> Dict[str, Union[float, bytes]]:
    """return Z-Score, model_quality_img_res, res_score_img_res in dict"""
    pdb_path = str(Path(pdb_path).resolve())
    b = browser or Browser()
    b.get('https://prosa.services.came.sbg.ac.at/prosa.php')
    btn = b.find_elements('//input[@type="file"]')[0]
    btn.send_keys(pdb_path)
    b.click(element='//input[@type="submit"]')
    xpath_result_line = '/html/body/div[2]/span[2]'
    if not b.wait_element(xpath_result_line, timeout=timeout):
        return put_err('Timeout', {})
    score = float(b.find_elements(xpath_result_line)[0].text)
    model_quality_img_link = b.find_elements('/html/body/div[2]/p/a')[0].get_attribute('href')
    model_quality_img_res = requests.get(model_quality_img_link)
    res_score_img_link = b.find_elements('/html/body/div[3]/p/span/a')[0].get_attribute('href')
    res_score_img_res = requests.get(res_score_img_link)
    return {'Z-Score': score,
            'model_quality_img_res': model_quality_img_res,
            'res_score_img_res': res_score_img_res}


def get_score_from_ModEval(modkey: str, pdb_path: str, align_file_path: str,
                           browser: Browser = None, timeout: int = 1800, **kwargs) -> Dict[str, Union[float, str]]:
    """return RMSD, overlap, identity, Z-DOPE in dict"""
    pdb_path = str(Path(pdb_path).resolve())
    align_file_path = str(Path(align_file_path).resolve())
    b = browser or Browser()
    b.get('https://modbase.compbio.ucsf.edu/evaluation/')
    b.send_key(key=modkey, element='//input[@name="modkey"]')
    btn = b.find_elements('//input[@type="file" and @name="pdb_file"]')[0]
    btn.send_keys(pdb_path)
    btn = b.find_elements('//input[@type="file" and @name="alignment_file"]')[0]
    btn.send_keys(align_file_path)
    b.click(element='//input[@type="submit"]')
    if not b.wait_element('//*[@id="resulttable"]/table/tbody/tr[7]/td[2]', timeout=timeout):
        return put_err('Timeout', {})
    rmsd_text = b.find_elements('//*[@id="resulttable"]/table/tbody/tr[7]/td[2]')[0].text.strip()
    rmsd = float(re.findall(r'[0-9.]+', rmsd_text)[0])
    overlap_text = b.find_elements('//*[@id="resulttable"]/table/tbody/tr[8]/td[2]')[0].text.strip()
    overlap = float(re.findall(r'[0-9.]+', overlap_text)[0])
    identity_text = b.find_elements('//*[@id="resulttable"]/table/tbody/tr[11]/td[2]')[0].text.strip()
    identity = float(re.findall(r'[0-9.]+', identity_text)[0])
    z_dope_text = b.find_elements('//*[@id="resulttable"]/table/tbody/tr[12]/td[2]')[0].text.strip()
    z_dope = float(re.findall(r'[0-9.]+', z_dope_text)[0])
    return {'RMSD': rmsd, 'overlap': overlap,
            'identity': identity, 'Z-DOPE': z_dope}


def get_score_from_MolProbity(pdb_path: str, browser: Browser = None, timeout: int = 1800, **kwargs) -> Dict[str, Union[float, str]]:
    """return Z-Score, Ramachandran Favorability, Ramachandran Outerness in dict"""
    pdb_path = str(Path(pdb_path).resolve())
    b = browser or Browser()
    b.get('http://molprobity.biochem.duke.edu/index.php')
    btn = b.find_elements('//input[@type="file" and @name="uploadFile"]')[0]
    btn.send_keys(pdb_path)
    b.click(element='//input[@type="submit" and @value="Upload >"]')
    if not b.wait_element('//input[@type="submit" and @value="Continue >"]', timeout=timeout):
        return put_err('Timeout', {})
    b.click(element='//input[@type="submit" and @value="Continue >"]')
    result_link = b.find_elements('/html/body/table/tbody/tr[2]/td[2]/div[1]/div/table/tbody/tr/td[2]/table/tbody/tr[4]/td[2]/a')[0].get_attribute('href')
    b.get(result_link)
    b.click(element='//input[@type="submit" and @value="Run programs to perform these analyses >"]')
    xpath_z_score = '/html/body/table/tbody/tr[2]/td/div/p[1]/table/tbody/tr[5]/td[2]'
    if not b.wait_element(xpath_z_score, timeout=timeout):
        return put_err('Timeout', {})
    z_score_text = b.find_elements(xpath_z_score)[0].text
    rama_favor_text = b.find_elements('/html/body/table/tbody/tr[2]/td/div/p[1]/table/tbody/tr[4]/td[3]')[0].text
    rama_outer_text = b.find_elements('/html/body/table/tbody/tr[2]/td/div/p[1]/table/tbody/tr[3]/td[3]')[0].text
    return {'Z-Score': z_score_text,
            'Ramachandran Favorability': rama_favor_text,
            'Ramachandran Outerness': rama_outer_text}


def get_score_from_ProQ3(pdb_path: str, browser: Browser = None, timeout: int = 1200, **kwargs) -> Dict[str, Union[float, str]]:
    """return ProQ2D, ProQRosCenD, ProQRosCenD, ProQ3D in dict"""
    pdb_path = str(Path(pdb_path).resolve())
    b = browser or Browser()
    b.get('https://proq3.bioinfo.se/pred/')
    btn = b.find_elements('//input[@type="file" and @name="modelfile"]')[0]
    btn.send_keys(pdb_path)
    b.click(element='//input[@type="submit" and @value="Submit"]')
    xpath_result_line = '//*[@id="jobtable"]/tbody'
    if not b.wait_element(xpath_result_line, timeout=timeout):
        return put_err('Timeout', {})
    z_score_text = b.find_elements(xpath_result_line)[0].text
    lines = z_score_text.split('\n')
    return {k:float(v) if re.findall('[0-9.]+', v) else v \
            for k,v in zip(lines[0].split(' ')[-4:], lines[1].split(' ')[-4:])}


def get_score_from_SAVES(pdb_path: str, browser: Browser = None, timeout: int = 1200, **kwargs) -> Dict[str, Union[float, str]]:
    """return ProQ2D, ProQRosCenD, ProQRosCenD, ProQ3D in dict"""
    pdb_path = str(Path(pdb_path).resolve())
    b = browser or Browser()
    b.get('https://saves.mbi.ucla.edu/')
    btn = b.find_elements('//input[@type="file" and @name="pdbfile"]')[0]
    btn.send_keys(pdb_path)
    b.click(element='//input[@type="submit" and @value="Run programs"]')
    if not b.wait_element('/html/body/h2[1]/a[1]', timeout=timeout):
        return put_err('Timeout', {})
    # start check progress
    b.click(element='//*[@id="errat"]')
    b.click(element='//*[@id="whatcheck"]')
    b.click(element='//*[@id="procheck"]')
    # wait for result
    if not b.wait_element(['//*[@id="werrat"]/div/center/center/h1',
                           '//*[@id="wwhatcheck"]/div/center',
                           '//*[@id="wprocheck"]/div/center'], timeout=timeout):
        return put_err('Timeout', {})
    errat_text = float(b.find_elements('//*[@id="werrat"]/div/center/center/h1')[0].text.strip())
    whatcheck_bad = len(b.find_elements('//*[@id="wwhatcheck"]/div/center//span[@class="wcitem bad"]'))
    whatcheck_mid = len(b.find_elements('//*[@id="wwhatcheck"]/div/center//span[@class="wcitem mid"]'))
    whatcheck_good = len(b.find_elements('//*[@id="wwhatcheck"]/div/center//span[@class="wcitem good"]'))
    procheck_errors = int(b.find_elements('//*[@id="wprocheck"]/div/center/ul/li[1]/span')[0].text.split(':')[-1])
    procheck_warnings = int(b.find_elements('//*[@id="wprocheck"]/div/center/ul/li[2]/span')[0].text.split(':')[-1])
    procheck_pass = int(b.find_elements('//*[@id="wprocheck"]/div/center/ul/li[3]/span')[0].text.split(':')[-1])
    # parse whatcheck result
    return {'errat': errat_text, 'whatcheck_bad': whatcheck_bad, 'whatcheck_mid': whatcheck_mid, 'whatcheck_good': whatcheck_good,
            'procheck_errors': procheck_errors, 'procheck_warnings': procheck_warnings, 'procheck_pass': procheck_pass}
    
    
def get_score_from_QMEANDisCo(pdb_path: str, browser: Browser = None, timeout: int = 1200, **kwargs):
    """return swissmodel score in dict"""
    b = browser or Browser()
    b.get('https://swissmodel.expasy.org/qmean/')
    btn = b.find_elements('//input[@id="structureFile" and @name="structureFile"]')[0]
    btn.send_keys(pdb_path)
    # wait upload and click submit
    if not b.wait_element('//*[@id="files"]/div', timeout=timeout):
        return put_err('Timeout', {})
    b.click(element='//*[@id="submitButton"]')
    # wait for result
    if not b.wait_element('//*[@id="main"]/div[2]/div[1]/div/span', timeout=timeout):
        return put_err('Timeout', {})
    score_text = b.find_elements('//*[@id="main"]/div[2]/div[1]/div/span')[0].text.strip()
    return {'QMEANDisCo Global': score_text.split(':')[-1]}
    
    
def get_score_from_QMEAN(pdb_path: str, browser: Browser = None, timeout: int = 1200, **kwargs):
    """return swissmodel score in dict"""
    b = browser or Browser()
    b.get('https://swissmodel.expasy.org/qmean/')
    btn = b.find_elements('//input[@id="structureFile" and @name="structureFile"]')[0]
    btn.send_keys(pdb_path)
    # wait upload and click submit
    if not b.wait_element('//*[@id="files"]/div', timeout=timeout):
        return put_err('Timeout', {})
    b.click(element='//*[@id="qmean_radio"]')
    b.click(element='//*[@id="submitButton"]')
    # wait for result
    if not b.wait_element('//*[@id="main"]/div[2]/div[1]/div/span', timeout=timeout):
        return put_err('Timeout', {})
    score_text = b.find_elements('//*[@id="main"]/div[2]/div[1]/div/span')[0].text.strip()
    return {'QMEAN4': score_text.split(':')[-1]}
    
    
_name2server = {
    'ProQ': get_score_from_proq,
    'VoroMQA': get_score_from_VoroMQA,
    'ProSA': get_score_from_ProSA,
    'ModEval': get_score_from_ModEval,
    'MolProbity': get_score_from_MolProbity,
    'ProQ3': get_score_from_ProQ3,
    'SAVES': get_score_from_SAVES,
    'QMEANDisCo': get_score_from_QMEANDisCo,
    'QMEAN': get_score_from_QMEAN,
}
    
    
def get_eval_info_from_servers(pdb_path: str, align_file_path: str = None,
                                modkey: str = None, browser: Browser = None,
                                servers: List[str] = None) -> Dict[str, Dict[str, Union[float, str, bytes]]]:
    """return a dict of evaluation results"""
    servers = servers or _name2server.keys()
    if not servers:
        return put_err('No server specified, return empty result', {})
    pdb_path = str(Path(pdb_path).resolve())
    b = browser or Browser()
    results = {}
    for server in servers:
        fn = _name2server[server]
        try:
            result = fn(pdb_path=pdb_path, align_file_path=align_file_path, modkey=modkey, browser=b)
        except Exception as e:
            result = {'Error': str(e)}
        results[server] = result
    return results


__all__ = [
    'get_score_from_proq',
    'get_score_from_VoroMQA',
    'get_score_from_ProSA',
    'get_score_from_ModEval',
    'get_score_from_MolProbity',
    'get_score_from_ProQ3',
    'get_score_from_SAVES',
    'get_score_from_QMEANDisCo',
    'get_score_from_QMEAN',
    '_name2server',
    'get_eval_info_from_web',
]


if __name__ == '__main__':
    print(get_score_from_proq('data_tmp/pdb/RECEPTOR.pdb'))
    print(get_score_from_VoroMQA('data_tmp/pdb/RECEPTOR.pdb'))
    print(get_score_from_ProSA('data_tmp/pdb/RECEPTOR.pdb'))
    # print(get_score_from_ModEval('data_tmp/pdb/RECEPTOR.pdb'))
    print(get_score_from_MolProbity('data_tmp/pdb/RECEPTOR.pdb'))
    print(get_score_from_ProQ3('data_tmp/pdb/RECEPTOR.pdb'))
    print(get_score_from_SAVES('data_tmp/pdb/RECEPTOR.pdb'))


from __future__ import absolute_import, print_function

import math
import re
import time
import tkinter as tk
import tkinter.filedialog as tkFileDialog
from typing import Any, Callable, List, Optional, Union

from mbapy_lite.base import put_err
from mbapy_lite.file import decode_bits_to_str, opts_file
from mbapy_lite.game import BaseInfo
from mbapy_lite.web_utils.task import TaskPool
from mbapy_lite.stats.cluster import KMeans, KMeansBackend

import numpy as np
from pymol import cmd
from tqdm import tqdm

from lazydock.utils import uuid4

# atom-type, atom-number, atom-name, residue-name, chain-name, residue-number, x, y, z, occupancy, temperature-factor
# ATOM      1  CA  LYS     7     136.747 133.408 135.880 -0.06 +0.10 OA
PDB_PATTERN = r"(ATOM|HETATM) +(\d+) +(\w+) +(\w+) +(\w+)? +(\d+) +([\d\-\.]+) +([\d\-\.]+) +([\d\-\.]+) +([\+\-][\d\-\.]+) +([\+\-][\d\-\.]+) [ \.\-\+\d]+ ([A-Z]+)"
PDB_FORMAT = "{:6s}{:>5s}  {:<3s} {:>3s} {:1s}{:>4s}    {:>8s}{:>8s}{:>8s}{:>6s}{:>6s}          {:>2s}  "
PDB_FORMAT2= "{:6s}{:>5s} {:<4s} {:>3s} {:1s}{:>4s}    {:>8s}{:>8s}{:>8s}{:>6s}{:>6s}          {:>2s}  "


def calcu_RMSD(pose1: Union[str, 'ADModel'], pose2: Union[str, 'ADModel'], _cmd = None):
    _cmd = _cmd or cmd
    def _get_pose(pose):
        if isinstance(pose, ADModel):
            name = uuid4()
            _cmd.read_pdbstr(pose.as_pdb_string(), name)
            return name, True
        return pose, False
    pml_pose1, is_loaded1 = _get_pose(pose1)
    pml_pose2, is_loaded2 = _get_pose(pose2)
    rmsd = _cmd.rms(pml_pose1, pml_pose2, cycles=0) # cutoff = float: outlier rejection cutoff (only if cycles>0) {default: 2.0}
    if is_loaded1:
        _cmd.delete(pml_pose1)
    if is_loaded2:
        _cmd.delete(pml_pose2)
    return rmsd


class ADModel(BaseInfo):
    """STORAGE CLASS FOR DOCKED LIGAND"""
    def __init__(self, content: str = None, _sort_atom_by_res: bool = False,
                 _parse2std: bool = False, default_chain: str = 'Z'):
        self.energy = 0.
        self.name = ''
        self.run_idx = 0
        self.info = ''
        self.pdb_string = ''
        self.default_chain = default_chain
        self.pdb_lines = []
        self.coords = None
        self.coords_only_bb = None
        if content is not None:
            self.parse_content(content, _sort_atom_by_res, _parse2std)

    def parse_content(self, content: str, _sort_atom_by_res: bool = False,
                      _parse2std: bool = False):
        # parse pdb lines
        self.info = content
        self.pdb_lines = list(map(lambda x: x[0], re.findall(r'((ATOM|HETATM).+?\n)', self.info)))
        self.pdb_atoms = list(map(list, re.findall(PDB_PATTERN, self.info)))
        if len(self.pdb_atoms) != len(self.pdb_lines):
            put_err(f'pdb_atoms and pdb_lines length not match: {len(self.pdb_atoms)} vs {len(self.pdb_lines)}, just use pdb_lines')
            _sort_atom_by_res = _parse2std = False
        if _sort_atom_by_res:
            pack = sorted(zip(self.pdb_lines, self.pdb_atoms), key = lambda x : (x[1][4], int(x[1][5]), int(x[1][1])))
            self.pdb_lines, self.pdb_atoms = zip(*pack)
        if _parse2std:
            for atom in self.pdb_atoms:
                if not atom[4]:
                    atom[4] = self.default_chain
                if len(atom[2]) == 4:
                    atom[-1] = atom[2][1] # such as 1HH1
                elif len(atom[2]) >= 1:
                    if atom[2][0].isdigit():
                        atom[-1] = atom[2][1] # such as 1HH
                    else:
                        atom[-1] = atom[2][0] # such as H, CA, NH1
            self.pdb_lines = [PDB_FORMAT2.format(*line) if len(line[2])==4 else PDB_FORMAT.format(*line) for line in self.pdb_atoms]
            self.pdb_string = '\n'.join(self.pdb_lines)
        else:
            self.pdb_string = ''.join(self.pdb_lines)
        # parse energy could be str.find?
        energy_line = re.findall(r'USER.+?Free Energy of Binding.+?\n', self.info)
        if energy_line:
            entr = energy_line[0].split('=')[1]
            self.energy = float(entr.split()[0])
        else:
            energy_line = re.findall(r'REMARK.+?VINA RESULT.+?\n', self.info)
            if energy_line:
                entr = energy_line[0].split(':')[1]
                self.energy = float(entr.split()[0])
            else:
                self.energy = None
        # parse run idx
        self.run_idx = int(re.findall(r'MODEL +(\d+)', self.info)[0])

    def as_pdb_string(self):
        return self.pdb_string

    def info_string(self):
        return self.info
    
    def parse_coords(self, only_bb: bool = True):
        if self.coords is None or only_bb != self.coords_only_bb:
            self.coords_only_bb = only_bb
            self.coords = np.array([list(map(float, line[6:9])) for line in self.pdb_atoms if line[2] in ['C', 'N', 'O', 'CA'] or not only_bb])
        return self.coords
    
    def calcu_rmsd_by_coords(self, other: 'ADModel', only_bb: bool = True):
        self_coords = self.parse_coords(only_bb)
        other_coords = other.parse_coords(only_bb)
        d2 = np.sum((self_coords - other_coords)**2, axis=1)
        return np.sqrt(np.mean(d2))
    
    def rmsd(self, other: 'ADModel', backend: str = 'pymol', only_bb: bool = True):
        if backend == 'pymol':
            return calcu_RMSD(self, other)
        elif backend == 'rdkit':
            from rdkit import Chem
            from rdkit.Chem import AllChem
            pose1 = Chem.MolFromPDBBlock(self.as_pdb_string(), removeHs=True, sanitize=False)
            pose2 = Chem.MolFromPDBBlock(other.as_pdb_string(), removeHs=True, sanitize=False)
            return AllChem.GetBestRMS(pose1, pose2)
        elif backend == 'numpy':
            return self.calcu_rmsd_by_coords(other, only_bb)
        else:
            raise ValueError(f'Unsupported backend: {backend}')


class RmsdClusterBackend(KMeansBackend):
    def __init__(self) -> None:
        super().__init__('scipy')
    def cdist(self, data, centers):
        return np.sqrt(np.mean(np.sum((data[:, None] - centers[None, :])**2, axis=-1), axis=-1))


class DlgFile(BaseInfo):
    def __init__(self, path: str = None, content: str = None,
                 sort_pdb_line_by_res: bool = False,
                 parse2std: bool = False):
        # load content from file if path is provided
        self.path = path
        self.sort_pdb_line_by_res = sort_pdb_line_by_res
        self.parse2std = parse2std
        if path is not None:
            self.content = decode_bits_to_str(opts_file(path, 'rb'))
        elif content is not None:
            self.content = content
        # decode content to pose_lst
        if self.content is not None:
            self.pose_lst: List[ADModel] = self.decode_content()
            self.sort_pose()
        else:
            self.pose_lst: List[ADModel] = []
        self.n2i = {}
        
    def __len__(self):
        return len(self.pose_lst)
        
    def sort_pose(self, key: Callable[[ADModel], Any] = None,
                  inplace: bool = True, reverse: bool = False) -> None:
        if key is None:
            key = lambda x : x.energy
        if inplace:
            self.pose_lst.sort(key=key, reverse=reverse)
            ret = self.pose_lst
        else:
            ret = sorted(self.pose_lst, key=key, reverse=reverse)
        return ret
        
    def decode_content(self):
        dlg_pose_lst = []
        for model in re.findall('MODEL.+?ENDMDL', self.content, re.DOTALL):
            model = model.replace('\nDOCKED: ', '\n')
            dlg_pose_lst.append(ADModel(model, self.sort_pdb_line_by_res, self.parse2std))
        return dlg_pose_lst
    
    def asign_pose_name(self, pose_names: List[str]):
        if len(pose_names) != len(self.pose_lst):
            raise ValueError("Number of pose names must match number of poses")
        for i, name in enumerate(pose_names):
            self.n2i[name] = i
        
    def asign_prop(self, prop: str, value: List[Any]):
        if value is not None and len(value) == len(self.pose_lst):
            setattr(self, prop, value)
                
    def set_pose_prop(self, prop: str, value: Any, pose_name: str = None, pose_idx: int = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        if not hasattr(self, prop):
            setattr(self, prop, [None]*len(self.pose_lst))
        getattr(self, prop)[pose_idx] = value
            
    def get_pose(self, pose_name: str = None, pose_idx: int = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        return self.pose_lst[pose_idx]
    
    def get_pose_prop(self, prop: str, pose_name: str = None, pose_idx: int = None, default: Any = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        if hasattr(self, prop):
            return getattr(self, prop)[pose_idx]
        else:
            return default
        
    def rmsd(self, backend: str = 'pymol', taskpool: TaskPool = None, verbose: bool = False):
        # if backend is numpy, use matrix calculation to calculate rmsd for acceleration
        if backend == 'numpy':
            # [N_pose, M_atom, D_3]
            coords = np.concatenate([[pose.parse_coords()] for pose in self.pose_lst], axis=0)
            # d2 = [N_pose, N_pose, M_atom, D_3] => [N_pose, N_pose, M_atom]: np.sum((coords[None, :, :] - coords[:, None, :])**2, axis=-1)
            # sum+devide => mean => [N_pose, N_pose, M_atom] => [N_pose, N_pose]
            return np.sqrt(np.mean(np.sum((coords[None, :, :] - coords[:, None, :])**2, axis=-1), axis=-1))
        # if is other backend, loop over all pairs and calculate rmsd
        rmsd_mat = np.zeros((len(self.pose_lst), len(self.pose_lst))).tolist()
        for i, pose_i in tqdm(enumerate(self.pose_lst), desc='Calculating RMSD', total=len(self.pose_lst), disable=not verbose):
            for j, pose_j in enumerate(self.pose_lst):
                if i > j:
                    continue
                if taskpool is None:
                    rmsd_mat[i][j] = pose_i.rmsd(pose_j, backend=backend)
                else:
                    while taskpool.count_waiting_tasks() > 0:
                        time.sleep(0.1)
                    rmsd_mat[i][j] = taskpool.add_task(f'{i}-{j}', pose_i.rmsd, pose_j, backend=backend)
        for i in range(len(rmsd_mat)):
            for j in range(len(rmsd_mat)):
                if i > j:
                    rmsd_mat[i][j] = rmsd_mat[j][i]
                elif taskpool is not None:
                    rmsd_mat[i][j] = taskpool.query_task(f'{i}-{j}', block=True, timeout=999)
        return rmsd_mat
    
    def rmsd_cluster(self, n_clusters: int):
        coords = np.concatenate([[pose.parse_coords()] for pose in self.pose_lst], axis=0)
        cluster = KMeans(n_clusters, backend=RmsdClusterBackend())
        return cluster.fit_predict(coords)
    
    def calcu_SSE_SSR(self, rmsd_mat: np.ndarray, groups_idx: np.ndarray,
                      centers_idx: Optional[np.ndarray] = None, center_idx: Optional[int] = None):
        """
        Parameters
            - rmsd_mat: [N_pose, N_pose], RMSD matrix return by DlgFile.rmsd
            - groups_idx: [N_pose], clustering result return by DlgFile.rmsd_cluster or other clustering algorithm
            - centers_idx: [N_cluster], clustering center index, **must in the rmsd matrix**, if not provided, use the minimum of each group
            - center_idx: int, center of the all poses, **must in the rmsd matrix**, if not provided, use the minimum of all poses
            
        Returns
            - SSE: 组内误差平方和
            - SSR: 组间误差平方和
            - SSR/SST: 
        """
        if centers_idx is None:
            centers_idx = np.array([np.argmin(rmsd_mat[groups_idx==i]) for i in np.unique(groups_idx)])
        if center_idx is None:
            center_idx = np.argmin(rmsd_mat)
        sse, ssr = 0, 0
        for i in np.unique(groups_idx):
            sse += np.sum(rmsd_mat[groups_idx==i, centers_idx[i]]**2)
            ssr += np.sum(groups_idx==i) * rmsd_mat[centers_idx[i], center_idx]**2
        return sse, ssr, ssr/(sse+ssr)


def tk_file_dialog_wrapper(*args, **kwargs):
    def ret_wrapper(tk_file_dialog_func):
        def core_wrapper(*args, **kwargs):
            parent = tk.Tk()
            result = tk_file_dialog_func(*args, parent = parent, **kwargs)
            parent.withdraw()
            if result == "":
                return None
            else:
                return result
        return core_wrapper
    return ret_wrapper
            

class MyFileDialog:
    def __init__(self, types=[("Executable", "*")], initialdir: str = None):
        self.initialdir = initialdir
        self.types = types

    @tk_file_dialog_wrapper()
    def get_open_file(self, parent):
        return tkFileDialog.askopenfilename(parent=parent, initialdir = self.initialdir, filetypes=self.types)


    @tk_file_dialog_wrapper()
    def get_save_file(self, parent):
        return tkFileDialog.asksaveasfilename(parent=parent, initialdir = self.initialdir, filetypes=self.types)
        
    @tk_file_dialog_wrapper()
    def get_ask_dir(self, parent):
        return tkFileDialog.askdirectory(parent=parent, initialdir = self.initialdir)


if __name__ == '__main__':
    # from mbapy_lite.base import TimeCosts
    # @TimeCosts(10)
    # def load_test(idx):
    #     DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=True)
    # load_test()
    normal = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=False, parse2std=False)
    std = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=False, parse2std=True)
    from pymol import cmd
    from rdkit import Chem
    cmd.reinitialize()
    cmd.read_pdbstr(std.pose_lst[0].as_pdb_string(), 'std')
    print(std.pose_lst[0].as_pdb_string(), '\n\n')
    print(cmd.get_pdbstr('std'))
    ligand = Chem.MolFromPDBBlock(cmd.get_pdbstr('std'), removeHs=True, sanitize=False)
    
    import seaborn as sns
    import matplotlib.pyplot as plt
    from mbapy_lite.stats import pca
    dlg = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=True, parse2std=True)
    dlg.sort_pose(inplace=True)
    energies = np.array([pose.energy for pose in dlg.pose_lst])
    rmsd = dlg.rmsd('numpy')
    
    # sns.clustermap(rmsd)
    # plt.show()
    
    np.random.seed(3407)
    rs = []
    for k in range(2, 7):
        groups_idx = dlg.rmsd_cluster(k)
        sse, ssr, r = dlg.calcu_SSE_SSR(rmsd, groups_idx)
        rs.append(r)
    plt.plot(list(range(2, 7)), rs)
    plt.show()
    
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    group_cmaps = ['Reds', 'Blues', 'Greens']
    groups_idx = dlg.rmsd_cluster(len(group_cmaps))
    dots = pca(rmsd, 3)
    for i, group_cmap in enumerate(group_cmaps):
        ax.scatter(dots[groups_idx==i, 0], dots[groups_idx==i, 1], dots[groups_idx==i, 2],
                   c=energies[groups_idx==i], cmap=group_cmap, alpha=0.4, s=40)
    plt.show()
    


'''
Date: 2024-10-11 10:33:10
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-10-11 19:33:45
Description: 
'''
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from plip.exchange.report import BindingSiteReport
from plip.structure.preparation import PDBComplex
from pymol import cmd

if __name__ == '__main__':
    from lazydock.pml.interaction_utils import sort_func
    from lazydock.utils import uuid4
else:
    from ..utils import uuid4
    from .interaction_utils import sort_func
    
    
def get_atom_level_interactions(mol, receptor_chain: str, ligand_chain: str, cutoff: float = 4.):
    """
    """
    interactions = {}
    for ligand, interaction in mol.interaction_sets.items():
        info = BindingSiteReport(interaction)
        for name, feat, values in [['Hydrophobic Interactions', info.hydrophobic_features, info.hydrophobic_info],
                                  ['Hydrogen Bonds', info.hbond_features, info.hbond_info],
                                  ['Water Bridges', info.waterbridge_features, info.waterbridge_info],
                                  ['Salt Bridges', info.saltbridge_features, info.saltbridge_info],
                                  ['pi-Stacking', info.pistacking_features, info.pistacking_info],
                                  ['pi-Cation Interactions', info.pication_features, info.pication_info],
                                  ['Halogen Bonds', info.halogen_features, info.halogen_info],
                                  ['Metal Complexes', info.metal_features, info.metal_info]]:
            interactions.setdefault(name, [])
            for value in values:
                find_idx_fn = lambda x: feat.index(list(filter(lambda y: x == y, feat))[0])
                rec_idx, lig_idx = find_idx_fn('RESNR'), find_idx_fn('RESNR_LIG')
                dist_term = 'DIST'
                if name == 'Hydrogen Bonds':
                    dist_term = 'DIST_H-A'
                elif name == 'Water Bridges':
                    dist_term = 'DIST_A-W'
                elif name == 'pi-Stacking':
                    dist_term = 'CENTDIST'
                dist_idx = find_idx_fn(dist_term)
                rec_res, lig_res, dist = value[rec_idx:rec_idx+3], value[lig_idx:lig_idx+3], float(value[dist_idx])
                if dist <= cutoff and rec_res[-1] == receptor_chain and lig_res[-1] == ligand_chain:
                    interactions[name].append((rec_res, lig_res, dist))
    return interactions


def merge_interaction_df(interaction: Dict[str, List[Tuple[Tuple[int, str, str], Tuple[int, str, str], float]]],
                         interaction_df: pd.DataFrame, distance_cutoff: float):
    """merge the interactions returned by calcu_atom_level_interactions to interaction_df."""
    # index format: CHAIN_ID:RESI:RESN
    for interaction_type, values in interaction.items():
        for single_inter in values:
            # single_inter: ((217, 'VAL', 'A'), (10, 'PHE', 'Z'), 3.71)
            receptor_res = f'{single_inter[0][2]}:{single_inter[0][0]}:{single_inter[0][1]}'
            ligand_res = f'{single_inter[1][2]}:{single_inter[1][0]}:{single_inter[1][1]}'
            points = distance_cutoff - single_inter[2]
            if ligand_res not in interaction_df.index or receptor_res not in interaction_df.columns:
                interaction_df.loc[ligand_res, receptor_res] = points
            elif np.isnan(interaction_df.loc[ligand_res, receptor_res]):
                interaction_df.loc[ligand_res, receptor_res] = points
            else:
                interaction_df.loc[ligand_res, receptor_res] += points
    return interaction_df


def calcu_receptor_poses_interaction(receptor: str, poses: List[str], cutoff: float = 4., **kwargs):
    """
    calcu interactions between one receptor and one ligand with many poses using PLIP-python.
    Parameters:
        - receptor (str): receptor pymol name
        - poses (List[str]):  ligand pymol names
        - cutoff (float): cutoff distance, default 4
        
    Returns:
        interactions (dict):
            - key: ligand name
            - value: interactions dict between receptor and ligand
                - key: interaction type, includes: Hydrophobic Interactions, Hydrogen Bonds, Water Bridges, Salt Bridges, pi-Stacking, pi-Cation Interactions, Halogen Bonds, Metal Complexes
                - value: list of interaction info: ((receptor_resi, receptor_resn, receptor_chain), (ligand_resi, ligand_resn, ligand_chain), distance)
        
        interaction_df (pd.DataFrame): , interactions between receptor and ligand, in the format of ligand-residue-residue matrix, with the value of each cell is the interaction score between two atoms.
            interaction_df.loc[ligand_res, receptor_res] = score
    """
    # prepare interactions
    receptor_chain = cmd.get_chains(receptor)[0]
    all_interactions, interaction_df = {}, pd.DataFrame()
    # calcu for each ligand
    for ligand in poses:
        ligand_chain = cmd.get_chains(ligand)[0]
        # calcu interaction
        sele_complex = uuid4()
        cmd.select(sele_complex, f'{ligand} or {receptor}')
        mol = PDBComplex()
        mol.load_pdb(cmd.get_pdbstr(sele_complex), as_string=True)
        mol.analyze()
        all_interactions[ligand] = get_atom_level_interactions(mol, receptor_chain, ligand_chain, cutoff)
        # merge interactions by res
        merge_interaction_df(all_interactions[ligand], interaction_df, cutoff)
    if not interaction_df.empty:
        # sort res
        interaction_df.sort_index(axis=0, inplace=True, key=sort_func)
        interaction_df.sort_index(axis=1, inplace=True, key=sort_func)
        interaction_df.fillna(0, inplace=True)
    else:
        return None, None
    return all_interactions, interaction_df


if __name__ == '__main__':
    # dev code
    from lazydock.pml.autodock_utils import DlgFile
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'RECEPTOR')
    dlg = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=True, parse2std=True)
    dlg.sort_pose()
    pose_lst = []
    for i, pose in enumerate(dlg.pose_lst[:10]):
        pose_lst.append(f'ligand_{i}')
        cmd.read_pdbstr(pose.as_pdb_string(), pose_lst[-1])
        cmd.alter(f'ligand_{i}', 'type="HETATM"')
    calcu_receptor_poses_interaction('RECEPTOR', pose_lst)


'''
Date: 2024-05-13 15:40:33
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-05-14 16:52:08
Description: 
'''
from typing import Union

import pyrosetta
from mbapy_lite.base import get_default_call_for_None, parameter_checker
from pyrosetta.rosetta.core.chemical import ChemicalManager
from pyrosetta.rosetta.core.conformation import ResidueFactory
from pyrosetta.rosetta.core.pose import (Pose, append_pose_to_pose,
                                         get_chain_from_chain_id,
                                         get_chain_id_from_chain,
                                         renumber_pdbinfo_based_on_conf_chains)
from pyrosetta.rosetta.protocols.grafting import delete_region
from pyrosetta.rosetta.protocols.simple_moves import (
    AddPDBInfoMover, DeleteChainMover, SwitchChainOrderMover,
    SwitchResidueTypeSetMover)
from pyrosetta.toolbox import mutate_residue


class _Pose(object):
    
    add_pdbinfo_mover = AddPDBInfoMover()
    delete_chain_mover = DeleteChainMover()
    swap_chain_mover = None
    chm = ChemicalManager.get_instance()
    rts = None
    to_centroid_mover = SwitchResidueTypeSetMover('centroid')
    to_fullatom_mover = SwitchResidueTypeSetMover('full_atom')
    
    def __init__(self, seq: str = None, pdb_path: str = None, rcsb: str = None) -> None:
        # init public attributes
        if self.swap_chain_mover is None:
            self.swap_chain_mover = SwitchChainOrderMover()
        if self.rts is None:
            self.rts = self.chm.residue_type_set("fa_standard")
        # init pose from seq or pdb
        if seq is not None:
            self.pose = pyrosetta.pose_from_sequence(seq)
        elif pdb_path is not None:
            self.pose = pyrosetta.pose_from_pdb(pdb_path)
        elif rcsb is not None:
            self.pose = pyrosetta.pose_from_rcsb(rcsb)
        else:
            self.pose = Pose()
        # init pdb_info from pose
        self.pdb_info = self.pose.pdb_info() or self.add_pdbinfo_mover.apply(self.pose)
            
    @parameter_checker(pos = lambda x: x in {'before', 'after'})
    def add_resi(self, resi_name3: str, insert_pos: int, pos: str = 'after'):
        new_rsd = ResidueFactory.create_residue(self.rts.name_map(resi_name3))
        if pos == 'before':
            self.pose.prepend_polymer_residue_before_seqpos(new_rsd, insert_pos, True)
        else:
            self.pose.append_polymer_residue_after_seqpos(new_rsd, insert_pos, True)
        renumber_pdbinfo_based_on_conf_chains(self.pose)
        return self
            
    def delete_resi(self, resi_id: int):
        self.pose.delete_polymer_residue(resi_id)
        renumber_pdbinfo_based_on_conf_chains(self.pose)
        return self
    
    def mutate_resi(self, resi_id: int, new_name1: str, pack_radius: float = 0.0):
        mutate_residue(self.pose, resi_id, new_name1, pack_radius)
            
    def add_chain(self, chain: Union['_Pose', Pose], jump_resi_id: int = None) -> '_Pose':
        jump_resi_id = get_default_call_for_None(jump_resi_id, self.pose.total_residue)
        chain = chain.pose if isinstance(chain, _Pose) else chain
        self.pose.append_pose_by_jump(chain, jump_anchor_residue=jump_resi_id)
        self.pose.update_residue_neighbors()
        self.pose.update_pose_chains_from_pdb_chains()
        self.pose.conformation().detect_disulfides()
        renumber_pdbinfo_based_on_conf_chains(self.pose)
        return self
    
    def del_region(self, start_resi_id: int, stop_resi_id: int):
        delete_region(self.pose, start_resi_id, stop_resi_id)
        renumber_pdbinfo_based_on_conf_chains(self.pose)
    
    def merge_chain(self, chain: Union['_Pose', Pose], new_chain: bool = False) -> '_Pose':
        if isinstance(chain, _Pose):
            chain = chain.pose
        append_pose_to_pose(self.pose, chain, new_chain)
        return self
    
    def delete_chain(self, chain_id: int = None, chain_name: str = None) -> '_Pose':
        if isinstance(chain_name, str):
            chain_id = self.get_chain_id(chain_name)
        elif not isinstance(chain_id, int):
            raise ValueError('Either chain_id or chain_name must be specified.')
        self.delete_chain_mover.chain_num(chain_id)
        self.delete_chain_mover.apply(self.pose)
        return self
        
    def swap_chain(self, order: str | list[str]) -> '_Pose':
        if isinstance(order, list) and not isinstance(order, str):
            order = ''.join(order)
        else:
            raise ValueError('order must be a str or list of chain names')
        self.swap_chain_mover.chain_order(order)
        self.swap_chain_mover.apply(self.pose)
        return self
    
    def split_chain(self, return_Pose: bool = False) -> list['_Pose'] | list[Pose]:
        chains = self.pose.split_by_chain()
        return [_Pose(chain) for chain in chains] if return_Pose else chains
    
    def get_resi_id_in_pose_via_pdb(self, chain_name: str, resi_id: int) -> int:
        return self.pdb_info.pdb2pose(chain_name, resi_id)
    
    def get_resi_id_in_pdb_via_pose(self, resi_id: int) -> tuple[str, int]:
        resi_info = self.pdb_info.pose2pdb(resi_id)
        resi_id, chain_name, _ = resi_info.split(' ') # '2 A ' -> ['2', 'A', '']
        return chain_name, int(resi_id)
    
    def get_chain_id(self, chain_name: str) -> int:
        return get_chain_id_from_chain(chain_name, self.pose)
    
    def get_chain_name(self, chain_id: int = None, resi_id: int = None) -> str:
        if chain_id is not None:
            return get_chain_from_chain_id(chain_id, self.pose)
        elif resi_id is not None:
            return self.pdb_info.chain(resi_id)
        else:
            raise ValueError('Either chain_id or resi_id must be specified.')
        
    def get_chain_seq(self, chain_id: int = None, chain_name: str = None) -> str:
        if isinstance(chain_id, int):
            return self.pose.chain_sequence(chain_id)
        elif isinstance(chain_name, str):
            return self.pose.chain_sequence(self.get_chain_id(chain_name))
        else:
            raise ValueError('Either chain_id or chain_name must be specified.')
        

__all__ = [
    '_Pose',
    'renumber_pdbinfo_based_on_conf_chains',
]


if __name__ == '__main__':
    pyrosetta.init()
    pose = _Pose(seq='AAAA')
    pose.get_resi_id_in_pdb_via_pose(2)


'''
Date: 2024-05-06 17:18:10
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-10-31 16:59:57
Description: 
'''

import json
import pathlib
import sys

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")
requirements = json.loads((here / "requirements.json").read_text())
version_info = (here / "lazydock/__version__.py").read_text(encoding="utf-8")
for line in version_info.split('\n'):
    if '__version__' in line:
        __version__ = line[line.find('"')+1:-1]
    if '__author_email__' in line:
        __author_email__ = line[line.find('"')+1:-1]
    if '__author__' in line:
        __author__ = line[line.find('"')+1:-1]
    if '__url__' in line:
        __url__ = line[line.find('"')+1:-1]
        
# decide which dyna lib to compare with
if sys.platform == "win32":
    dynlib = [
    ]
elif sys.platform in ["linux", "linux2"]:
    dynlib = [
    ]
else:
    dynlib = [
    ]
    

setup(
    name = "lazydock",
    version = __version__,

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        # "Programming Language :: Python :: 3 :: Only",
    ],
        
    keywords = ["molecular docking", "Utilities"],
    description = "A Python package for molecular docking",
    long_description = long_description,
    long_description_content_type='text/markdown',
    python_requires=">=3.7, <3.13",
    license = "MIT Licence",

    url = __url__,
    author = __author__,
    author_email = __author_email__,
    
    packages = find_packages(exclude=["test", "test."]),
    include_package_data = True, # define in MANIFEST.in file
    package_data = {"lazydock": dynlib},
    
    entry_points={
        "console_scripts": [
            # "lazydock=lazydock.scripts:main",
        ],
    },
    
    platforms = "any",
    
    install_requires=requirements['std'],
)

# pip install .

# python setup.py sdist
# twine upload dist/lazydock-0.5.0.tar.gz



import gzip
import os
import pickle
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from mbapy_lite.base import put_err
from mbapy_lite.file import decode_bits_to_str, opts_file
from mbapy_lite.plot import save_show
from nicegui import ui
from pymol import api, cmd

from lazydock.pml.autodock_utils import DlgFile, MyFileDialog
from lazydock.pml.interaction_utils import \
    calcu_receptor_poses_interaction as calc_pml_interaction
from lazydock.pml.interaction_utils import filter_interaction_df

try:
    from lazydock.pml.plip_interaction import \
        calcu_receptor_poses_interaction as calc_plip_interaction
except ImportError:
    # if plip is not installed, use interaction_utils instead
    put_err('plip is not installed, just use pymol as plip instead')
    from lazydock.pml.interaction_utils import \
        calcu_receptor_poses_interaction as calc_plip_interaction

from lazydock.utils import uuid4


class LazyPose:
    """load dlg to pose, show pose, save pose"""
    def __init__(self, app):
        self._app = app
        self._app.ui_update_func.append(self.ui_update_ui)
        # pose storage
        self.dlg_pose: Dict[str, DlgFile] = {}
        self.now_dlg_name = None
        self.last_pose_name = None
        self.now_pose_name = None
        # show pose control
        self.sort_pdb_by_res = True
        self.parse2std = True
        self.show_best_per = 10
        self.save_lig_with_HETATM = True
        self.save_with_each_header = False
        self.save_original_pdbstring = False
        # select receptor
        self.ui_molecule = None
        self.ui_sele = None
        self.sele_molecule = None
        self.sele_selection = None
        
    def ui_update_ui(self):
        if self.ui_sele is not None:
            self.ui_sele.set_options(self._app._now_selection)
        if self._app.lazy_dlg.pose_page.dlg_pose:
            dlg_pose: Dict[str, DlgFile] = self._app.lazy_dlg.pose_page.dlg_pose
            all_ligands = [dlg_pose[dlg].get_pose_prop('pml_name', lig) for dlg in dlg_pose for lig in dlg_pose[dlg].n2i if dlg_pose[dlg].get_pose_prop('is_show', lig, default = False)]
            self.ui_molecule.set_options(list(set(self._app._now_molecule) - set(all_ligands)))
        else:
            self.ui_molecule.set_options(self._app._now_molecule)
        
    def build_gui(self):
        """called by LazyDLG.__init__"""
        with ui.row().classes('w-full'):
            ui.upload(label = 'Load DLG', multiple=True, auto_upload=True,
                    on_upload=self.load_dlg_file, on_multi_upload=self.load_dlg_file).props('no-caps')
            ui.button('refresh GUI', on_click=self.ui_make_dlg_file_list.refresh).props('no-caps')
            # load and save control
            with ui.column():
                ui.checkbox('sort pdb by res', value=self.sort_pdb_by_res).bind_value_to(self, 'sort_pdb_by_res')
                ui.checkbox('parse2std', value=self.parse2std).bind_value_to(self, 'parse2std')
                ui.checkbox('save lig with HETATM', value=self.save_lig_with_HETATM).bind_value_to(self,'save_lig_with_HETATM')
                ui.checkbox('save with each header', value=self.save_with_each_header).bind_value_to(self, 'save_with_each_header')
                ui.checkbox('save orignal pdbstring', value=self.save_original_pdbstring).bind_value_to(self,'save_original_pdbstring')
            # sele receptor
            with ui.card().classes('w-1/2'):
                with ui.row().classes('w-full'):
                    ui.label('select a receptor')
                    ui.button('save showed complex', on_click=self.save_showed_complex).props('no-caps').bind_enabled_from(self, 'now_dlg_name')
                with ui.row().classes('w-full'):
                    self.ui_molecule = ui.select(self._app._now_molecule,
                                                label = 'select a molecule').bind_value_to(self, 'sele_molecule').classes('w-2/5').props('use-chips')
                    # ui.label('OR').classes('w-1/5')
                    self.ui_sele = ui.select(self._app._now_selection,
                                            label = 'select a selection').bind_value_to(self, 'sele_selection').classes('w-2/5').props('use-chips')
        self.ui_make_dlg_file_list()
        
    @ui.refreshable
    def ui_make_dlg_file_list(self):
        with ui.row().classes('w-full h-full'):
            # file list
            with ui.tabs(value = self.now_dlg_name).props('vertical active-bg-color=blue').classes('w-1/8 h-2/3') as file_tabs:
                for n in self.dlg_pose:
                    ui.tab(n).props('no-caps').classes('w-full').tooltip(n)
            file_tabs.on('click', self.handle_dlg_tab_click)
            if not self.now_dlg_name:
                return file_tabs
            # pose list
            self.now_pose_name = list(self.dlg_pose[self.now_dlg_name].n2i.keys())[0]
            with ui.tabs(value = self.now_pose_name).props('vertical').classes('w-1/8 h-2/3') as pose_tabs:
                for n in self.dlg_pose[self.now_dlg_name].n2i:
                    text_col = 'text-blue' if self.dlg_pose[self.now_dlg_name].get_pose_prop('is_show', n) else 'text-black'
                    ui_tab = ui.tab(n).classes(f'w-full {text_col} no-caps').tooltip(n).on('click', self.handle_pose_tab_click)
                    self.dlg_pose[self.now_dlg_name].set_pose_prop('ui_tab', ui_tab, n)
                        
            pose_tabs.bind_value_to(self, 'now_pose_name')
            # pose info
            with ui.column().classes('w-3/5'):
                ui.label().props('no-cap').bind_text_from(self, 'now_dlg_name', lambda x: f'DLG: {x}')
                # pose control
                with ui.row().classes('w-full'):
                    ui.button('show best 10', on_click=self.show_best_10_poses).props('no-caps').classes('flex flex-grow')
                    ui.number('best percentage', min=0, max=100, value=self.show_best_per).bind_value_to(self, 'show_best_per').classes('flex flex-grow')
                    ui.button('show percentage', on_click=self.show_percentage_poses).props('no-caps').classes('flex flex-grow')
                    ui.button('show all', on_click=self.show_all_poses).props('no-caps').classes('flex flex-grow')
                    ui.button('hide all', on_click=self.hide_all_poses).props('no-caps').classes('flex flex-grow')
                    ui.button('save showed', on_click=self.save_showed_pose).props('no-caps').classes('flex flex-grow')
                # pose info
                with ui.card().classes('w-full h-full'):
                    self.build_pose_info_gui()
                    
    def handle_dlg_tab_click(self, event):
        click_dlg = event.sender._props['model-value']
        self.now_dlg_name = click_dlg
        self.now_pose_name = list(self.dlg_pose[click_dlg].n2i.keys())[0]
        self.ui_make_dlg_file_list.refresh()
                    
    def handle_pose_tab_click(self, event):
        click_pose = event.sender._props['name']
        if self.dlg_pose[self.now_dlg_name].get_pose_prop('is_show', click_pose):
            self.now_pose_name = None
            self.hide_pose(self.now_dlg_name, click_pose)
        else:
            self.now_pose_name = click_pose
            self.show_pose(self.now_dlg_name, click_pose)
        self.build_pose_info_gui.refresh()
                    
    @ui.refreshable
    def build_pose_info_gui(self):
        ui.label().props('no-cap').bind_text_from(self, 'now_pose_name', lambda x: f'Pose: {x}')
        if self.now_dlg_name and self.now_pose_name:
            pml_name = self.dlg_pose[self.now_dlg_name].get_pose_prop('pml_name', self.now_pose_name)
            energy = self.dlg_pose[self.now_dlg_name].get_pose(self.now_pose_name).energy
            ui.label(f'{pml_name} docked Energy: {energy:10.4f} kcal/mol')
            ui.textarea().bind_value_from(self, 'now_dlg_name', lambda x: self.dlg_pose[x].get_pose(self.now_pose_name).info_string()).classes('w-full h-full flex flex-grow')
        
    async def load_dlg_file(self, event):
        if hasattr(event, 'name'):
            names, contents = [event.name], [event.content]
        else:
            names, contents = event.names, event.contents
        for file_name, content in zip(names, contents):
            dlg_name = Path(file_name).stem
            if dlg_name in self.dlg_pose:
                continue # TODO: don't kown why, will load again on same file, cause read null at the second time
            dlg_content = decode_bits_to_str(content.read())
            self.dlg_pose[dlg_name] = DlgFile(content=dlg_content,
                                              sort_pdb_line_by_res=self.sort_pdb_by_res,
                                              parse2std=self.parse2std)
            self.dlg_pose[dlg_name].sort_pose()
            self.dlg_pose[dlg_name].asign_pose_name(list(map(lambda x: f'{dlg_name}::{x+1}',
                                                             range(len(self.dlg_pose[dlg_name].pose_lst)))))
            self.dlg_pose[dlg_name].asign_prop('is_show', [False] * len(self.dlg_pose[dlg_name].pose_lst))
            self.dlg_pose[dlg_name].asign_prop('pml_name', list(map(lambda x: x.replace(' ', '_').replace('::', '_'),
                                                                       self.dlg_pose[dlg_name].n2i.keys())))
            self.now_dlg_name = dlg_name
        self.ui_make_dlg_file_list.refresh()
        
    def show_pose(self, dlg_name: str, pose_name: str):
        """
            - set self.ui_dlg_pose[dlg_name]['is_pose_show'][pose_name] to True
            - create pymol obj neamed pose_name.replace(' ', '_').replace('::', '_')
            - show the obj in pymol in sticks representation
        """
        self.dlg_pose[dlg_name].set_pose_prop('is_show', True, pose_name)
        ui_tab = self.dlg_pose[dlg_name].get_pose_prop('ui_tab', pose_name)
        ui_tab.classes(add='text-blue', remove='text-black')
        view = cmd.get_view()
        pose = self.dlg_pose[dlg_name].get_pose(pose_name)
        cmd.read_pdbstr(pose.as_pdb_string(), pose_name)
        pml_name = self.dlg_pose[dlg_name].get_pose_prop('pml_name', pose_name)
        cmd.show('sticks', pml_name)
        cmd.set_view(view)
        cmd.zoom(pml_name)

    def show_all_poses(self):
        for pose_name in list(self.dlg_pose[self.now_dlg_name].n2i.keys()):
            self.show_pose(self.now_dlg_name, pose_name)

    def show_best_10_poses(self, n = 10):
        for pose_name in list(self.dlg_pose[self.now_dlg_name].n2i.keys())[:n]:
            self.show_pose(self.now_dlg_name, pose_name)
            
    def show_percentage_poses(self):
        n = int(len(self.dlg_pose[self.now_dlg_name].n2i) * self.show_best_per / 100)
        self.show_best_10_poses(n)
        
    def hide_pose(self, dlg_name: str, pose_name: str):
        if self.dlg_pose[dlg_name].get_pose_prop('is_show', pose_name):
            ui_tab = self.dlg_pose[dlg_name].get_pose_prop('ui_tab', pose_name)
            ui_tab.classes(add='text-black', remove='text-blue')
            self.dlg_pose[dlg_name].set_pose_prop('is_show', False, pose_name)
            cmd.delete(self.dlg_pose[dlg_name].get_pose_prop('pml_name', pose_name))

    def hide_all_poses(self):
        for pose_name in self.dlg_pose[self.now_dlg_name].n2i:
            self.hide_pose(self.now_dlg_name, pose_name)
            
    def save_showed_pose(self):
        if any(self.dlg_pose[self.now_dlg_name].get_pose_prop('is_show', pose_name)\
                for pose_name in self.dlg_pose[self.now_dlg_name].n2i):
            save_dir = MyFileDialog().get_ask_dir()
            if not save_dir:
                ui.notify('Please select a directory to save')
                return None
        else:
            ui.notify('Please show at least one pose')
            return None
        for pose_name in self.dlg_pose[self.now_dlg_name].n2i:
            if self.dlg_pose[self.now_dlg_name].get_pose_prop('is_show', pose_name):
                pml_name = self.dlg_pose[self.now_dlg_name].get_pose_prop('pml_name', pose_name)
                if self.save_lig_with_HETATM:
                    cmd.alter(pml_name, 'type="HETATM"')
                if self.save_original_pdbstring:
                    pose = self.dlg_pose[self.now_dlg_name].get_pose(pose_name)
                    opts_file(os.path.join(save_dir, f'{pml_name}.pdb'), 'w', data = pose.as_pdb_string())
                else:
                    cmd.save(os.path.join(save_dir, f'{pml_name}.pdb'), pml_name)
                print(f'{pml_name} saved to {save_dir}')

    def delete_dlg(self):
        cmd.delete(self.now_dlg_name + '_*')
        del self.dlg_pose[self.now_dlg_name]
        self.now_dlg_name = self.now_dlg_name[list(self.dlg_pose.keys())[0]] if self.now_dlg_name else None
        
    def save_showed_complex(self):
        if self.sele_molecule is None and self.sele_selection is None:
            ui.notify('Please select a receptor and a selection first')
            return None
        receptor = self.sele_molecule or self.sele_selection
        save_dir = MyFileDialog().get_ask_dir()
        if not save_dir:
            ui.notify('Please select a directory to save')
            return None
        for name in self.dlg_pose[self.now_dlg_name].n2i:
            if self.dlg_pose[self.now_dlg_name].get_pose_prop('is_show', name):
                pml_name = self.dlg_pose[self.now_dlg_name].get_pose_prop('pml_name', name)
                if self.save_lig_with_HETATM:
                    cmd.alter(pml_name, 'type="HETATM"')
                pdb_path = os.path.join(save_dir, f'{receptor}_{pml_name}.pdb')
                if self.save_with_each_header:
                    api.multisave(pdb_path, receptor, append = 0)
                    api.multisave(pdb_path, pml_name, append = 1)
                else:
                    tmp_sele = uuid4()
                    cmd.select(tmp_sele, f'{receptor} or {pml_name}')
                    cmd.save(pdb_path, tmp_sele)
                    cmd.delete(tmp_sele)
                print(f'{receptor} and {pml_name} saved to {pdb_path}')
                
                
class ResultPage(LazyPose):
    """plot energy curve and RMSD heatmap for pose in dlg"""
    def __init__(self, app):
        self._app = app
        self._app.ui_update_func.append(self.ui_update_ui)
        self.energy_fig = None
        self.rmsd_fig = None
        self.energy_data = None
        self.rmsd_data = None
        # select receptor
        self.ui_dlg = None
        self.sele_dlg = None
        
    def ui_update_ui(self):
        self.ui_dlg.set_options(list(self._app.lazy_dlg.pose_page.dlg_pose.keys()))
        
    def calculate_energy_curve(self):
        if self.sele_dlg is None:
            return ui.notify('Please select a DLG')
        dlg_pose: DlgFile = self._app.lazy_dlg.pose_page.dlg_pose[self.sele_dlg]
        idx_energy = [[pose.run_idx, pose.energy] for pose in dlg_pose.pose_lst]
        sorted_energies_by_idx = list(map(lambda x: x[1], sorted(idx_energy, key=lambda x: x[0])))
        sorted_energies = sorted(sorted_energies_by_idx, reverse=True)
        pooled_energies_by_idx = [min(sorted_energies_by_idx[:i]) for i in range(1, len(sorted_energies_by_idx)+1)]
        self.energy_data = (sorted_energies_by_idx, sorted_energies, pooled_energies_by_idx)
        ui.notify(f'Energy curve calculated for {len(dlg_pose.pose_lst)} poses, [{min(sorted_energies_by_idx)} ~ {max(sorted_energies_by_idx)}]')
        
    @ui.refreshable
    def plot_energy_curve(self):
        plt.close(self.energy_fig)
        self.energy_fig = None
        if self.energy_data is not None:
            sorted_energies_by_idx, sorted_energies, pooled_energies_by_idx = self.energy_data
            mean_energy = np.mean(sorted_energies)
            with ui.pyplot(figsize=(12, 8), close=False) as fig:
                self.energy_fig = fig.fig
                gs = fig.fig.add_gridspec(1, 2,  width_ratios=(4, 1), left=0.1, right=0.9, bottom=0.1, top=0.9,
                      wspace=0.05, hspace=0.05)
                # Create the Axes.
                ax = self.energy_fig.add_subplot(gs[0, 0])
                ax_histy = self.energy_fig.add_subplot(gs[0, 1], sharey=ax)
                # Draw the scatter plot and marginals.
                ax.plot(pooled_energies_by_idx, linewidth=2, label='Minimum Energy in Run Order')
                ax.plot([mean_energy] * len(pooled_energies_by_idx), c='black', linewidth=2, label=f'Mean Energy: {mean_energy:.4f} kcal/mol')
                ax.scatter(list(range(len(sorted_energies_by_idx))), sorted_energies_by_idx,
                           alpha=0.4, c='green', s=50, label='Individual Energy in Run Order')
                ax.scatter(list(range(len(sorted_energies))), sorted_energies,
                           alpha=0.2, c='red', s=30, label='Individual Energy in Descending Order')
                ax.set_xlabel('Pose Index', fontdict={'size': 14})
                ax.set_ylabel('Energy (kcal/mol)', fontdict={'size': 14})
                ax.set_title(f'Energy Curve for {self.sele_dlg}', fontdict={'size': 16})
                # plot hist
                ax_histy.tick_params(axis="y", labelleft=False)
                ax_histy.hist(sorted_energies, bins=int(max(sorted_energies)-min(sorted_energies)), orientation='horizontal')
                # minor works
                ax.tick_params(axis='both', which='major', labelsize=12)
                ax_histy.tick_params(axis='both', which='minor', labelsize=12)
                ax.legend(fontsize=12)
                plt.tight_layout()
        
    def build_gui(self):
        with ui.splitter(value = 20).classes('w-full h-full') as splitter:
            with splitter.before:
                # choose ligand
                with ui.card().classes('w-full'):
                    with ui.column().classes('w-full'):
                        ui.label('select a ligand').classes('w-full')
                        self.ui_dlg = ui.select(self._app.lazy_dlg.pose_page.now_dlg_name or [],
                                                label = 'select a dlg').bind_value_to(self, 'sele_dlg').classes('w-full').props('use-chips')
            # vitualization
            with splitter.after:
                with ui.row().classes('w-full'):
                    ui.button('calculate energy curve', on_click=self.calculate_energy_curve).props('no-caps')
                    ui.button('plot energy curve', on_click=self.plot_energy_curve.refresh).props('no-caps')
                self.plot_energy_curve()


class InteractionPage(LazyPose):
    def __init__(self, app):
        self._app = app
        self._app.ui_update_func.append(self.ui_update_ui)
        # select receptor
        self.ui_molecule = None
        self.ui_sele = None
        self.ui_dlg = None
        self.sele_molecule = None
        self.sele_selection = None
        self.sele_dlg = None
        # interaction
        self.interaction_method = 'pymol'
        self.interactions = None
        self.interaction_df = None
        self.interaction_mode = 0
        self.distance_cutoff = 4
        self.nagetive_factor = -0.5
        # vitulazation control
        self.fig = None
        self.align_vlim = False
        self.plot_cluster = False
        self.use_one_letter = True
        self.min_ligand_interaction = 0 # abs
        self.min_receptor_interaction = 0 # abs
        self.fig_w = 12
        self.fig_h = 7
        
    def ui_update_ui(self):
        super().ui_update_ui()
        self.ui_dlg.set_options(list(self._app.lazy_dlg.pose_page.dlg_pose.keys()))
            
    def calculate_interaction(self):
        # get receptor
        if self.sele_molecule is None and self.sele_selection is None:
            ui.notify('Please select a receptor')
            return None
        # get ligand
        if self.sele_dlg is None:
            ui.notify('Please select a DLG')
        ligands = []
        for name in self._app.lazy_dlg.pose_page.dlg_pose[self.sele_dlg].n2i:
            if self._app.lazy_dlg.pose_page.dlg_pose[self.sele_dlg].get_pose_prop('is_show', name):
                ligands.append(self._app.lazy_dlg.pose_page.dlg_pose[self.sele_dlg].get_pose_prop('pml_name', name))
        if not ligands:
            ui.notify('Please select at least one ligand')
            return None
        # calculate interactions
        if self.interaction_method == 'pymol':
            calc_fn = calc_pml_interaction
        elif self.interaction_method == 'PLIP':
            calc_fn = calc_plip_interaction
        else:
            assert False, f'Unknown interaction method: {self.interaction_method}'
        self.interactions, self.interaction_df = calc_fn(self.sele_molecule or self.sele_selection, ligands,
                                                         mode=self.interaction_mode, cutoff=self.distance_cutoff,
                                                         nagetive_factor=self.nagetive_factor)
        if self.interactions is not None:
            ui.notify(f'Interactions calculated for {len(ligands)} ligands')
        else:
            ui.notify(f'No interactions found between {self.sele_molecule or self.sele_selection} and {len(ligands)} ligands')
        return self.interactions
            
    @ui.refreshable
    def plot_interaction(self):
        plt.close(self.fig)
        self.fig = None
        if (self.interaction_df is not None) and (not self.interaction_df.empty):
            
            tmp_interaction_df = filter_interaction_df(self.interaction_df, self.min_receptor_interaction,
                                                       self.min_ligand_interaction, inplace=False)
            # plot
            with ui.pyplot(figsize=(self.fig_w, self.fig_h), close=False) as fig:
                vmax, vmin = tmp_interaction_df.max().max(), tmp_interaction_df.min().min()
                if self.align_vlim:
                    vlim = max(abs(vmax), abs(vmin))
                    vmax, vmin = vlim, -vlim
                if self.plot_cluster:
                    grid = sns.clustermap(tmp_interaction_df, cmap='coolwarm', figsize=(self.fig_w, self.fig_h),
                                        annot=True, fmt='.2f', vmax=vmax, vmin=vmin,
                                        linewidths=0.5, linecolor='black', cbar_kws={"shrink": 0.5})
                    self.fig = fig.fig = grid._figure
                else:
                    self.fig = fig.fig
                    grid = sns.heatmap(tmp_interaction_df, cmap='coolwarm', ax = fig.fig.gca(),
                                    annot=True, fmt='.2f', vmax=vmax, vmin=vmin,
                                    linewidths=0.5, linecolor='black', cbar_kws={"shrink": 0.5})
                plt.tight_layout()
        
    def build_gui(self):
        with ui.splitter(value = 20).classes('w-full h-full') as splitter:
            with splitter.before:
                # choose receptor
                with ui.card().classes('w-full'):
                    with ui.column().classes('w-full'):
                        ui.label('select a receptor')
                        self.ui_molecule = ui.select(self._app._now_molecule,
                                                    label = 'select a molecule').bind_value_to(self, 'sele_molecule').classes('w-full').props('use-chips')
                        self.ui_sele = ui.select(self._app._now_selection,
                                                label = 'select a selection').bind_value_to(self, 'sele_selection').classes('w-full').props('use-chips')
                # choose ligand
                with ui.card().classes('w-full'):
                    with ui.column().classes('w-full'):
                        ui.label('select a ligand').classes('w-full')
                        self.ui_dlg = ui.select(self._app.lazy_dlg.pose_page.now_dlg_name or [],
                                                label = 'select a dlg').bind_value_to(self, 'sele_dlg').classes('w-full').props('use-chips')
                # interaction calculation config
                with ui.card().classes('w-full'):
                    with ui.column().classes('w-full'):
                        ui.label('interaction calculation config').classes('w-full')
                        ui.select(['pymol', 'PLIP'], value=self.interaction_method).bind_value_to(self, 'interaction_method')
                        ui.number('distance cutoff', value=self.distance_cutoff).bind_value_to(self, 'distance_cutoff')
                        ui.select(options = list(range(9)), value = 0).bind_value_to(self, 'interaction_mode')
                        ui.number('nagetive factor', value=self.nagetive_factor).bind_value_to(self, 'nagetive_factor')
                # plot config
                with ui.card().classes('w-full'):
                    with ui.column().classes('w-full'):
                        ui.label('plot config').classes('w-full')
                        ui.checkbox('plot cluster', value=self.plot_cluster).bind_value_to(self, 'plot_cluster')
                        ui.checkbox('align vlim', value=self.align_vlim).bind_value_to(self, 'align_vlim')
                        ui.checkbox('use one letter', value=self.use_one_letter).bind_value_to(self, 'use_one_letter')
                        ui.number('min ligand interaction (abs)', value=self.min_ligand_interaction).bind_value_to(self,'min_ligand_interaction')
                        ui.number('min receptor interaction (abs)', value=self.min_receptor_interaction).bind_value_to(self,'min_receptor_interaction')
                        ui.number('fig width', value=self.fig_w).bind_value_to(self, 'fig_w')
                        ui.number('fig height', value=self.fig_h).bind_value_to(self, 'fig_h')
            # interaction vitualization
            with splitter.after:
                with ui.row():
                    ui.label('Interaction: ')
                    ui.button('calculate', on_click=self.calculate_interaction).classes('flex flex-grow').props('no-caps')
                    ui.button('save calcu result', on_click=self.save_calcu_result).classes('flex flex-grow').props('no-caps')
                    ui.button('plot', on_click=self.plot_interaction.refresh).classes('flex flex-grow').props('no-caps')
                    ui.button('save Plot', on_click=self.save_plot).classes('flex flex-grow').props('no-caps')
                self.plot_interaction()
        
    def save_calcu_result(self):
        if self.interaction_df is None:
            ui.notify('Please calculate the interaction first')
            return None
        data_path = MyFileDialog(types = [('LazyInteraction file', '*.pkl')],
                                 initialdir=os.getcwd()).get_save_file()
        if data_path is None:
            ui.notify('Please select a file to save')
            return None
        if not data_path.endswith('.pkl'):
            data_path += '.pkl'
        with open(data_path, 'wb') as f:
            f.write(gzip.compress(pickle.dumps(self.interactions)))            
            self.interaction_df.to_pickle(f, compression={'method': 'gzip', 'compresslevel': 1, 'mtime': 1})
        ui.notify(f'Interactions saved to {data_path}')
    
    def save_plot(self):
        if self.fig is None:
            ui.notify('Please plot the interaction first')
            return None
        fig_path = MyFileDialog(initialdir=os.getcwd()).get_save_file()
        if fig_path is None:
            ui.notify('Please select a file to save')
            return None
        save_show(fig_path, dpi = 600, show = False)
    

class LazyDLG:
    def __init__(self, app, _dev_mode: bool = False):
        self._app = app
        
        self.pose_page = LazyPose(self._app)
        self.result_page = ResultPage(self._app)
        self.analysis_page = InteractionPage(self._app)
        
    def build_gui(self):
        with ui.tabs().classes('w-full').props('align=left active-bg-color=blue') as tabs:
            self.ui_loader_tab = ui.tab('DLG Pose Loader').props('no-caps')
            self.ui_result_tab = ui.tab('DLG Pose Result').props('no-caps')
            self.ui_analysis_tab = ui.tab('DLG Pose Analysis').props('no-caps')
        with ui.tab_panels(tabs, value=self.ui_loader_tab).classes('w-full'):
            with ui.tab_panel(self.ui_loader_tab):
                self.pose_page.build_gui()
            with ui.tab_panel(self.ui_result_tab):
                self.result_page.build_gui()
            with ui.tab_panel(self.ui_analysis_tab):
                self.analysis_page.build_gui()
        # return self
        return self
        
        
# dev mode
if __name__ in {"__main__", "__mp_main__"}:
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'receptor')
    
    from main import GUILauncher
    GUILauncher()
    


'''
Date: 2024-05-06 17:19:43
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-10-31 16:59:27
Description: 
'''

__title__ = "lazydock"
__description__ = "Lazy Dock"
__url__ = "https://github.com/BHM-Bob/LazyDock"
__version__ = "0.5.0"
__build__ = "20241031_0"
__author__ = "BHM-Bob G"
__author_email__ = "bhmfly@foxmail.com"
__license__ = "MIT"